"""Layer for :py:class:`~tfts.models.autoformer`"""

from typing import Dict, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import AveragePooling1D, Conv1D, Dense, Dropout


class MovingAvg(tf.keras.layers.Layer):
    """
    Moving average block to highlight the trend of time series
    """

    def __init__(self, kernel_size: int, stride: int = 1):
        super().__init__()
        if kernel_size % 2 != 1:
            raise ValueError("Moving average kernel size must be an odd number")
        self.kernel_size = kernel_size
        self.stride = stride

    def build(self, input_shape: Tuple[Optional[int], ...]):
        super().build(input_shape)
        self.avg = AveragePooling1D(pool_size=self.kernel_size, strides=self.stride, padding="valid")

    def call(self, inputs):
        """
        Perform moving average for sequence

        Args:
            inputs: Input tensor.

        Returns:
            Output tensor.
        """
        front = tf.tile(inputs[:, :1, :], [1, (self.kernel_size - 1) // 2, 1])
        end = tf.tile(inputs[:, -1:, :], [1, (self.kernel_size - 1) // 2, 1])
        x = tf.concat([front, inputs, end], axis=1)
        x = self.avg(x)
        return x


class SeriesDecomp(tf.keras.layers.Layer):
    def __init__(self, kernel_size: int, name=None) -> None:
        super().__init__(name=name)
        self.kernel_size = kernel_size

    def build(self, input_shape: Tuple[Optional[int], ...]):
        super().build(input_shape)
        self.moving_avg = MovingAvg(self.kernel_size, stride=1)

    def call(self, x: tf.Tensor):
        """
        Perform time-series decomposition on the input tensor.

        Parameters
        ----------
        x : tf.Tensor
            A 3D tensor with shape (batch_size, sequence_length, input_dim).

        Returns
        -------
        Tuple[tf.Tensor, tf.Tensor]
            A tuple of two 3D tensors:
            - The residual tensor, shape (batch_size, sequence_length, input_dim).
            - The moving average tensor, which is a smoothed version of the input tensor.
        """
        moving_mean = self.moving_avg(x)
        trend = x - moving_mean
        return trend, moving_mean

    def get_config(self):
        config = {
            "kernel_size": self.kernel_size,
        }
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class AutoCorrelation(tf.keras.layers.Layer):
    """Self-Attention layer that computes time-delayed autocorrelation between queries and keys"""

    def __init__(self, d_model: int, num_attention_heads: int, attention_probs_dropout_prob: float = 0.0) -> None:
        super().__init__()
        if d_model % num_attention_heads != 0:
            raise ValueError(f"Hidden size {d_model} must be divisible by the number of heads {num_attention_heads}.")
        self.d_model = d_model
        self.num_attention_heads = num_attention_heads
        self.hidden_size = d_model // num_attention_heads
        self.attention_probs_dropout_prob = attention_probs_dropout_prob

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.wq = Dense(self.d_model, name="q")
        self.wk = Dense(self.d_model, name="k")
        self.wv = Dense(self.d_model, name="v")
        self.drop = Dropout(self.attention_probs_dropout_prob)
        self.dense = Dense(self.d_model, name="project")
        super().build(input_shape)

    def time_delay_agg(self, q, k, v):  # TODO: v not used in process
        """Compute time-delayed autocorrelation between queries and keys.

        Parameters
        ----------
        q : Tensor of shape (batch_size, num_attention_heads, time_steps, hidden_size)
            Queries.
        k : Tensor of shape (batch_size, num_attention_heads, time_steps, hidden_size)
            Keys.
        v : Tensor of shape (batch_size, num_attention_heads, time_steps, hidden_size)
            Values.

        Returns
        -------
        Tensor of shape (batch_size, num_attention_heads, time_steps, hidden_size)
            Time-delayed autocorrelation between queries and keys.
        """
        batch_size = tf.shape(q)[0]
        time_steps = tf.shape(q)[2]
        q_fft = tf.signal.rfft(tf.transpose(q, perm=[0, 1, 3, 2]))
        k_fft = tf.signal.rfft(tf.transpose(k, perm=[0, 1, 3, 2]))
        S_qk = q_fft * tf.math.conj(k_fft)
        R_qk = tf.signal.irfft(S_qk)

        init_index = tf.reshape(tf.range(time_steps), (1, 1, 1, -1))
        init_index = tf.tile(init_index, [batch_size, self.num_attention_heads, self.hidden_size, 1])
        top_k = int(2 * tf.math.log(tf.cast(time_steps, tf.float32)))
        # mean_value = tf.reduce_mean(R_qk, axis=1)
        weights, indices = tf.math.top_k(R_qk, top_k)

        tmp_corr = tf.nn.softmax(weights, axis=-1)
        # tmp_corr = tf.reshape(tmp_corr, (batch_size, 1, 32, -1))

        tmp_values = tf.tile(tf.transpose(q, perm=[0, 1, 3, 2]), tf.constant([1, 1, 1, 2]))
        delays_agg = tf.zeros_like(tf.transpose(q, perm=[0, 1, 3, 2]))

        for i in range(top_k):
            pattern = tf.gather(tmp_values, init_index + tf.expand_dims(indices[..., i], -1), axis=-1, batch_dims=-1)
            delays_agg = delays_agg + pattern * (tf.expand_dims(tmp_corr[..., i], axis=-1))
        return delays_agg

    def split_heads(self, x, batch_size):
        x = tf.reshape(x, (batch_size, -1, self.num_attention_heads, self.hidden_size))
        return tf.transpose(x, perm=[0, 2, 1, 3])  # (batch_size, num_attention_heads, timesteps, depth)

    def call(self, q, k, v, dynamic=True):
        """_summary_

        Parameters
        ----------
        q : _type_
            _description_
        k : _type_
            _description_
        v : _type_
            _description_
        dynamic : bool, optional
            _description_, by default True

        Returns
        -------
        _type_
            _description_
        """
        batch_size = tf.shape(q)[0]

        q = self.drop(self.wq(q))
        k = self.drop(self.wk(k))
        v = self.drop(self.wv(v))

        q = self.split_heads(q, batch_size)
        k = self.split_heads(k, batch_size)
        v = self.split_heads(v, batch_size)

        L = tf.shape(q)[2]
        S = tf.shape(v)[2]

        if tf.math.greater(L, S):
            zeros = tf.zeros_like(q[:, :, : (L - S), :])
            v = tf.concat([v, zeros], axis=2)
            k = tf.concat([k, zeros], axis=2)
        else:
            v = v[:, :, :L, :]
            k = k[:, :, :L, :]

        delays_agg = self.time_delay_agg(q, k, v)
        delays_agg = tf.transpose(delays_agg, [0, 3, 1, 2])
        concat_delays_agg = tf.reshape(delays_agg, (batch_size, -1, self.d_model))
        output = self.dense(concat_delays_agg)
        return output



"""Layer for :py:class:`~tfts.models.transformer`"""

from typing import Any, Callable, Dict, List, Optional, Tuple

import tensorflow as tf
from tensorflow.keras import activations, constraints, initializers, regularizers



"""
`Sequence to Sequence Learning with Neural Networks
<https://arxiv.org/abs/1409.3215>`_
"""

import logging
from typing import Optional

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import GRU, LSTM, Dense, GRUCell, LSTMCell

from tfts.layers.attention_layer import Attention

from .base import BaseConfig, BaseModel

logger = logging.getLogger(__name__)


class Seq2seqConfig(BaseConfig):
    model_type: str = "seq2seq"

    def __init__(
        self,
        rnn_hidden_size=64,
        rnn_type="gru",
        bi_direction=False,
        dense_hidden_size=64,
        num_stacked_layers=1,
        scheduled_sampling=0,
        use_attention=False,
        attention_size=64,
        num_attention_heads=2,
        attention_probs_dropout_prob=0,
    ):
        super(Seq2seqConfig, self).__init__()
        self.rnn_hidden_size = rnn_hidden_size
        self.rnn_type = rnn_type
        self.bi_direction = bi_direction
        self.dense_hidden_size = dense_hidden_size
        self.num_stacked_layers = num_stacked_layers
        self.scheduled_sampling = scheduled_sampling  # 0: teacher forcing
        self.use_attention = use_attention
        self.attention_size = attention_size
        self.num_attention_heads = num_attention_heads
        self.attention_probs_dropout_prob = attention_probs_dropout_prob

        if self.use_attention:
            assert self.attention_size == self.dense_hidden_size


class Seq2seq(BaseModel):
    """Seq2seq model for time series prediction with configurable encoder-decoder architectures."""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[Seq2seqConfig] = None):
        super(Seq2seq, self).__init__()
        self.config = config or Seq2seqConfig()
        self.predict_sequence_length = predict_sequence_length

        self.encoder = Encoder(
            rnn_size=self.config.rnn_hidden_size,
            rnn_type=self.config.rnn_type,
            dense_size=self.config.dense_hidden_size,
        )
        self.decoder = DecoderV1(
            rnn_size=self.config.rnn_hidden_size,
            rnn_type=self.config.rnn_type,
            predict_sequence_length=predict_sequence_length,
            use_attention=self.config.use_attention,
            attention_size=self.config.attention_size,
            num_attention_heads=self.config.num_attention_heads,
            attention_probs_dropout_prob=self.config.attention_probs_dropout_prob,
        )

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """Forward pass of the Seq2seq model.

        :param inputs: Input tensor.
        :param teacher: Ground truth for teacher forcing.
        :param return_dict: Whether to return outputs in a dict format.
        :return: Decoder outputs.
        """

        x, encoder_feature, decoder_feature = self._prepare_3d_inputs(inputs, ignore_decoder_inputs=False)
        encoder_outputs, encoder_state = self.encoder(encoder_feature)

        decoder_outputs = self.decoder(
            decoder_feature,
            decoder_init_input=x[:, -1, 0:1],
            init_state=encoder_state,
            teacher=teacher,
            scheduled_sampling=self.config.scheduled_sampling,
            encoder_output=encoder_outputs,
        )

        return decoder_outputs


class Encoder(tf.keras.layers.Layer):
    def __init__(self, rnn_size, rnn_type="gru", rnn_dropout=0, dense_size=32, return_state=False, **kwargs):
        super().__init__(**kwargs)
        self.rnn_type = rnn_type.lower()
        self.return_state = return_state
        if rnn_type == "gru":
            self.rnn = GRU(
                units=rnn_size, activation="tanh", return_state=True, return_sequences=True, dropout=rnn_dropout
            )
        elif rnn_type == "lstm":
            self.rnn = LSTM(
                units=rnn_size,
                activation="tanh",
                return_state=True,
                return_sequences=True,
                dropout=rnn_dropout,
            )
        else:
            raise ValueError(f"No supported RNN type: {rnn_type}")

        self.dense = Dense(units=dense_size, activation="tanh")

    def call(self, inputs):
        """Process input through the encoder RNN and dense layers.

        :param inputs: 3D Input tensor with shape (batch_size, seq_len, num_features)
        :return: Encoder outputs and state.

        outputs : tf.Tensor
            (batch_size, input_sequence_length, rnn_size)
        state : tf.Tensor or tuple of tf.Tensor
            Processed state(s) from the RNN:
            - For GRU: (batch_size, dense_size)
            - For LSTM: tuple of (batch_size, dense_size), (batch_size, dense_size)
        """
        if self.rnn_type == "gru":
            outputs, state = self.rnn(inputs)
            state = self.dense(state)
        elif self.rnn_type == "lstm":
            outputs, state_h, state_c = self.rnn(inputs)
            state_h = self.dense(state_h)
            state_c = self.dense(state_c)
            state = (state_h, state_c)
        else:
            raise ValueError(f"No supported rnn type of {self.rnn_type}")
        # encoder_hidden_state = tuple(self.dense(hidden_state) for _ in range(config['num_stacked_layers']))
        # outputs = self.dense(outputs)  # => batch_size * input_seq_length * dense_size
        return outputs, state


class DecoderV1(tf.keras.layers.Layer):
    def __init__(
        self,
        rnn_size=32,
        rnn_type="gru",
        predict_sequence_length=3,
        use_attention=False,
        attention_size=32,
        num_attention_heads=1,
        attention_probs_dropout_prob=0.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.predict_sequence_length = predict_sequence_length
        self.use_attention = use_attention
        self.rnn_type = rnn_type.lower()
        self.rnn_size = rnn_size
        self.attention_size = attention_size
        self.num_attention_heads = num_attention_heads
        self.attention_probs_dropout_prob = attention_probs_dropout_prob

    def build(self, input_shape):
        if self.rnn_type == "gru":
            self.rnn_cell = GRUCell(self.rnn_size)
        elif self.rnn_type == "lstm":
            self.rnn_cell = LSTMCell(units=self.rnn_size)
        else:
            raise ValueError(f"No supported rnn type of {self.rnn_type}")

        self.dense = Dense(units=1, activation=None)
        if self.use_attention:
            self.attention = Attention(
                hidden_size=self.attention_size,
                num_attention_heads=self.num_attention_heads,
                attention_probs_dropout_prob=self.attention_probs_dropout_prob,
            )
        super().build(input_shape)

    def call(
        self,
        decoder_features,
        decoder_init_input,
        init_state,
        teacher=None,
        scheduled_sampling=0,
        training=None,
        **kwargs,
    ):
        """Seq2seq decoder with attention mechanism.

        :param decoder_features: Decoder input features.
        :param decoder_init_input: Initial input for the decoder.
        :param init_state: Initial state from the encoder.
        :param teacher: Ground truth for teacher forcing.
        :param scheduled_sampling: Probability of using teacher forcing.
        :param training: Whether the model is in training mode.
        :return: Decoder output.
        """
        decoder_outputs = []
        prev_output = decoder_init_input
        prev_state = init_state
        if teacher is not None:
            teacher = tf.squeeze(teacher, 2)
            teachers = tf.split(teacher, self.predict_sequence_length, axis=1)

        for i in range(self.predict_sequence_length):
            if training:
                p = np.random.uniform(low=0, high=1, size=1)[0]
                if teacher is not None and p > scheduled_sampling:
                    this_input = teachers[i]
                else:
                    this_input = prev_output
            else:
                this_input = prev_output

            if decoder_features is not None:
                this_input = tf.concat([this_input, decoder_features[:, i]], axis=-1)

            if self.use_attention:
                if self.rnn_type.lower() == "gru":
                    # q: (batch, 1, feature), att_output: (batch, 1, feature)
                    att = self.attention(
                        tf.expand_dims(prev_state, 1), k=kwargs["encoder_output"], v=kwargs["encoder_output"]
                    )
                    att = tf.squeeze(att, 1)  # (batch, feature)
                elif self.rnn_type.lower() == "lstm":
                    # q: (batch, 1, feature * 2), att_output: (batch, 1, feature)
                    att = self.attention(
                        tf.expand_dims(tf.concat(prev_state, 1), 1),
                        k=kwargs["encoder_output"],
                        v=kwargs["encoder_output"],
                    )
                    att = tf.squeeze(att, 1)  # (batch, feature)

                this_input = tf.concat([this_input, att], axis=-1)

            this_output, this_state = self.rnn_cell(this_input, prev_state)
            prev_state = this_state
            prev_output = self.dense(this_output)
            decoder_outputs.append(prev_output)

        decoder_outputs = tf.concat(decoder_outputs, axis=-1)
        return tf.expand_dims(decoder_outputs, -1)


class DecoderV2(tf.keras.layers.Layer):
    def __init__(
        self,
        rnn_size=32,
        rnn_type="gru",
        predict_sequence_length=3,
        use_attention=False,
        attention_sizes=32,
        num_attention_heads=1,
        attention_probs_dropout_prob=0.0,
    ):
        super(DecoderV2, self).__init__()
        self.rnn_type = rnn_type
        self.rnn_size = rnn_size
        self.predict_sequence_length = predict_sequence_length
        self.use_attention = use_attention
        self.attention_sizes = attention_sizes
        self.num_attention_heads = num_attention_heads
        self.attention_probs_dropout_prob = attention_probs_dropout_prob

    def build(self, input_shape):
        if self.rnn_type.lower() == "gru":
            self.rnn_cell = GRUCell(self.rnn_size)
        elif self.rnn_type.lower() == "lstm":
            self.rnn = LSTMCell(units=self.rnn_size)
        self.dense = Dense(units=1)
        if self.use_attention:
            self.attention = Attention(
                hidden_size=self.attention_sizes,
                num_attention_heads=self.num_attention_heads,
                attention_probs_dropout_prob=self.attention_probs_dropout_prob,
            )
        super().build(input_shape)

    def forward(
        self,
        decoder_feature,
        decoder_init_value,
        init_state,
        teacher=None,
        scheduled_sampling=0,
        training=None,
        **kwargs,
    ):
        def cond_fn(time, prev_output, prev_state, decoder_output_ta):
            return time < self.predict_sequence_length

        def body(time, prev_output, prev_state, decoder_output_ta):
            if time == 0 or teacher is None:
                this_input = prev_output

            else:
                this_input = teacher[:, time - 1, :]

            if decoder_feature is not None:
                this_feature = decoder_feature[:, time, :]
                this_input = tf.concat([this_input, this_feature], axis=1)

            if self.use_attention:
                attention_feature = self.attention(
                    tf.expand_dims(prev_state[-1], 1), k=kwargs["encoder_output"], v=kwargs["encoder_output"]
                )
                attention_feature = tf.squeeze(attention_feature, 1)
                this_input = tf.concat([this_input, attention_feature], axis=-1)

            this_output, this_state = self.rnn_cell(this_input, prev_state)
            project_output = self.dense(this_output)
            decoder_output_ta = decoder_output_ta.write(time, project_output)
            return time + 1, project_output, this_state, decoder_output_ta

        loop_init = [
            tf.constant(0, dtype=tf.int32),  # steps
            decoder_init_value,  # decoder each step
            init_state,  # state
            tf.TensorArray(dtype=tf.float32, size=self.predict_sequence_length),
        ]
        _, _, _, decoder_outputs_ta = tf.while_loop(cond_fn, body, loop_init)

        decoder_outputs = decoder_outputs_ta.stack()
        decoder_outputs = tf.transpose(decoder_outputs, [1, 0, 2])
        return decoder_outputs

    def call(
        self,
        decoder_feature,
        decoder_init_input,
        init_state,
        teacher=None,
        scheduled_sampling=0,
        training=None,
        **kwargs,
    ):
        """Decoder model2

        Parameters
        ----------
        decoder_feature : _type_
            _description_
        init_state : _type_
            _description_
        decoder_init_input : _type_
            _description_
        teacher : _type_, optional
            _description_, by default None

        Returns
        -------
        _type_
            _description_
        """
        return self.forward(
            decoder_feature=decoder_feature,
            decoder_init_value=decoder_init_input,
            init_state=[init_state],  # for tf2
            teacher=teacher,
        )



"""
`U-Net: Convolutional Networks for Biomedical Image Segmentation
<https://arxiv.org/abs/1505.04597>`_
"""

from typing import Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Activation, AveragePooling1D, Concatenate, Conv1D, Lambda, UpSampling1D

from tfts.layers.unet_layer import conv_br, re_block

from .base import BaseConfig, BaseModel


class UnetConfig(BaseConfig):
    model_type: str = "unet"

    def __init__(
        self,
        units: int = 64,
        kernel_size: int = 2,
        depth: int = 2,
        pool_sizes: Tuple[int, int] = (2, 4),
        upsampling_factors: Tuple[int, int, int] = (4, 2, 2),
    ):
        super(UnetConfig, self).__init__()
        self.units = units
        self.kernel_size = kernel_size
        self.depth = depth
        self.pool_sizes = pool_sizes
        self.upsampling_factors = upsampling_factors


class Unet(BaseModel):
    """Unet model for sequence-to-sequence prediction tasks."""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[UnetConfig] = None):
        super(Unet, self).__init__()
        self.config = config or UnetConfig()
        self.predict_sequence_length = predict_sequence_length

        self.avg_pool1 = AveragePooling1D(pool_size=self.config.pool_sizes[0])
        self.avg_pool2 = AveragePooling1D(pool_size=self.config.pool_sizes[1])
        self.encoder = Encoder(units=self.config.units, kernel_size=self.config.kernel_size, depth=self.config.depth)
        self.decoder = Decoder(
            upsampling_factors=self.config.upsampling_factors,
            units=self.config.units,
            kernel_size=self.config.kernel_size,
            predict_seq_length=predict_sequence_length,
        )

    def __call__(
        self,
        x: tf.Tensor,
        training: bool = True,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """Forward pass through the model.

        Args:
            x: Input tensor of shape (batch_size, sequence_length, num_features).
            training: Boolean flag for training mode.

        Returns:
            Tensor: Output predictions.
        """
        pool1 = self.avg_pool1(x)
        pool2 = self.avg_pool2(x)

        encoder_output = self.encoder([x, pool1, pool2])
        decoder_outputs = self.decoder(encoder_output)

        return decoder_outputs


class Encoder(tf.keras.layers.Layer):
    """Encoder component for the Unet model."""

    def __init__(self, units: int = 64, kernel_size: int = 2, depth: int = 2):
        super().__init__()
        self.units = units
        self.kernel_size = kernel_size
        self.depth = depth

    def call(self, inputs: tf.Tensor):
        """Forward pass through the encoder.

        Args:
            inputs: Tuple containing the input tensor and pooled tensors.

        Returns:
            Tuple: Encoder outputs.
        """
        x, pool1, pool2 = inputs

        x = conv_br(x, self.units, self.kernel_size, 1, 1)  # => batch_size * sequence_length * units
        for i in range(self.depth):
            x = re_block(x, self.units, self.kernel_size, 1, 1)
        out_0 = x  # => batch_size * sequence_length * units

        x = conv_br(x, self.units * 2, self.kernel_size, 2, 1)
        for i in range(self.depth):
            x = re_block(x, self.units * 2, self.kernel_size, 1, 1)
        out_1 = x  # => batch_size * sequence/2 * units*2

        x = Concatenate()([x, pool1])
        x = conv_br(x, self.units * 3, self.kernel_size, 2, 1)
        for i in range(self.depth):
            x = re_block(x, self.units * 3, self.kernel_size, 1, 1)
        out_2 = x  # => batch_size * sequence/2, units*3

        x = Concatenate()([x, pool2])
        x = conv_br(x, self.units * 4, self.kernel_size, 4, 1)
        for i in range(self.depth):
            x = re_block(x, self.units * 4, self.kernel_size, 1, 1)
        return [out_0, out_1, out_2, x]


class Decoder(tf.keras.layers.Layer):
    """Decoder component for the Unet model."""

    def __init__(self, upsampling_factors, units: int = 64, kernel_size: int = 2, predict_seq_length: int = 1):
        super().__init__()
        self.upsampling_factors = upsampling_factors
        self.units = units
        self.kernel_size = kernel_size
        self.predict_seq_length = predict_seq_length

    def call(self, inputs: Tuple[tf.Tensor, tf.Tensor, tf.Tensor, tf.Tensor]) -> tf.Tensor:
        """Forward pass through the decoder.

        Args:
            inputs: Tuple containing encoder outputs.

        Returns:
            Tensor: Decoder output.
        """
        out_0, out_1, out_2, x = inputs

        x = UpSampling1D(self.upsampling_factors[0])(x)
        x = Concatenate()([x, out_2])
        x = conv_br(x, self.units * 3, self.kernel_size, 1, 1)

        x = UpSampling1D(self.upsampling_factors[1])(x)
        x = Concatenate()([x, out_1])
        x = conv_br(x, self.units * 2, self.kernel_size, 1, 1)

        x = UpSampling1D(self.upsampling_factors[2])(x)
        x = Concatenate()([x, out_0])
        x = conv_br(x, self.units, self.kernel_size, 1, 1)

        x = Conv1D(1, kernel_size=self.kernel_size, strides=1, padding="same")(x)
        x = Activation("sigmoid")(x)

        x = Lambda(lambda x: 12 * x)(x)
        # Todo: just a tricky way to change the batch*input_seq*1 -> batch_out_seq*1, need a more general way for time
        x = AveragePooling1D(pool_size=1, strides=4)(x)

        return x



"""
`BERT: Pre-training of Deep Bidirectional Transformers for Language Understanding
<https://arxiv.org/abs/1810.04805>`_
"""

import logging
from typing import Dict, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Dense, Reshape

from tfts.layers.embed_layer import DataEmbedding, TokenEmbedding
from tfts.models.transformer import Encoder

from .base import BaseConfig, BaseModel

logger = logging.getLogger(__name__)


class BertConfig(BaseConfig):

    model_type: str = "bert"

    def __init__(
        self,
        hidden_size: int = 64,
        num_layers: int = 2,
        num_attention_heads: int = 4,
        ffn_intermediate_size: int = 256,
        pooling_method: str = "mean",
        hidden_act: str = "gelu",
        hidden_dropout_prob: float = 0.0,
        attention_probs_dropout_prob: float = 0.0,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        pad_token_id: int = 0,
        positional_type: str = "positional encoding",
        use_cache: bool = False,
        dense_units: Tuple[int] = (512, 1024),
        classifier_dropout: Optional[float] = None,
        **kwargs: Dict[str, object],
    ) -> None:
        """Configuration class for BERT model, inheriting from BaseConfig.

        Args:
            hidden_size: The size of the hidden layers. Default is 64.
            num_hidden_layers: The number of hidden layers in the transformer encoder. Default is 2.
            num_attention_heads: The number of attention heads in each attention layer. Default is 4.
            ffn_intermediate_size: The size of the intermediate (feed-forward) layer. Default is 256.
            hidden_act: The activation function for hidden layers. Default is "gelu".
            hidden_dropout_prob: The dropout probability for hidden layers. Default is 0.1.
            attention_probs_dropout_prob: The dropout probability for attention probabilities. Default is 0.1.
            type_vocab_size: The vocabulary size for token types (usually 2). Default is 2.
            initializer_range: The standard deviation for weight initialization. Default is 0.02.
            layer_norm_eps: The epsilon value for layer normalization. Default is 1e-12.
            pad_token_id: The ID for the padding token. Default is 0.
            positional_type: The type of position embedding ("absolute" or "relative"). Default is "absolute".
            use_cache: Whether to use the cache during inference. Default is True.
            classifier_dropout: Dropout probability for the classifier layer. Default is None.
            **kwargs: Additional keyword arguments passed to the parent `BaseConfig` class.
        """

        super().__init__(**kwargs)

        self.hidden_size: int = hidden_size
        self.num_layers: int = num_layers
        self.num_attention_heads: int = num_attention_heads
        self.ffn_intermediate_size: int = ffn_intermediate_size
        self.pooling_method: str = pooling_method
        self.hidden_act: str = hidden_act
        self.hidden_dropout_prob: float = hidden_dropout_prob
        self.attention_probs_dropout_prob: float = attention_probs_dropout_prob
        self.type_vocab_size: int = type_vocab_size
        self.initializer_range: float = initializer_range
        self.layer_norm_eps: float = layer_norm_eps
        self.positional_type: str = positional_type
        self.use_cache: bool = use_cache
        self.dense_units: Tuple[int] = dense_units
        self.classifier_dropout: Optional[float] = classifier_dropout
        self.pad_token_id: int = pad_token_id


class Bert(BaseModel):
    """Bert model for time series forecasting.

    This model implements a transformer-based architecture (BERT) adapted for time series data.
    It processes time series inputs through a transformer encoder and produces predictions
    for future time steps.

    Parameters
    ----------
    predict_sequence_length : int, optional
        Number of future time steps to predict, by default 1
    config : BertConfig, optional
        Configuration parameters for the model, by default None

    Attributes
    ----------
    config : BertConfig
        Configuration object containing model hyperparameters
    predict_sequence_length : int
        Number of future time steps to predict
    encoder_embedding : DataEmbedding
        Embedding layer for encoder inputs
    encoder : Encoder
        Transformer encoder module
    dense_layers : List[Dense]
        List of dense layers for final projection
    """

    def __init__(self, predict_sequence_length: int = 1, config: Optional[BertConfig] = None) -> None:
        super(Bert, self).__init__(predict_sequence_length=predict_sequence_length, config=config)
        self.config = config or BertConfig()
        self.predict_sequence_length = predict_sequence_length
        self.built = False

    def build(self):
        """Builds the model layers with the input shape."""

        self.encoder_embedding = DataEmbedding(self.config.hidden_size, positional_type=self.config.positional_type)
        self.encoder = Encoder(
            num_hidden_layers=self.config.num_layers,
            hidden_size=self.config.hidden_size,
            num_attention_heads=self.config.num_attention_heads,
            attention_probs_dropout_prob=self.config.attention_probs_dropout_prob,
            ffn_intermediate_size=self.config.ffn_intermediate_size,
            hidden_dropout_prob=self.config.hidden_dropout_prob,
        )
        logger.debug(
            f"Created encoder with {self.config.num_layers} layers and {self.config.num_attention_heads} heads"
        )

        self.dense_layers = [
            Dense(unit, activation="relu", name=f"dense_{i}") for i, unit in enumerate(self.config.dense_units)
        ]
        logger.debug(f"Created {len(self.dense_layers)} dense layers with units: {self.config.dense_units}")

        self.projection = Dense(self.predict_sequence_length, activation="linear", name="projection")
        self.reshape = Reshape((self.predict_sequence_length, 1))
        logger.debug("Model building completed")

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> tf.Tensor:
        """Forward pass for Bert model.

        Parameters
        ----------
        inputs : Union[tf.Tensor, Tuple[tf.Tensor, ...], Dict[str, tf.Tensor]]
            Input tensor(s) - can be a single tensor, tuple of tensors, or dictionary
        teacher : tf.Tensor, optional
            Teacher forcing for autoregression, by default None
        output_hidden_states : bool, optional
            Whether to return hidden states, by default False
        return_dict : bool, optional
            Whether to return a dictionary of outputs, by default False

        Returns
        -------
        Union[tf.Tensor, Dict[str, tf.Tensor]]
            Model outputs - either a tensor of predictions or a dictionary of tensors
        """
        if not self.built:
            logger.info("Model not built yet, building now...")
            self.build()
            self.built = True

        try:
            x, encoder_feature, _ = self._prepare_3d_inputs(inputs)
            encoder_feature = self.encoder_embedding(encoder_feature)
            memory = self.encoder(encoder_feature, mask=None)

            if output_hidden_states:
                # (batch_size, train_sequence_length, hidden_size)
                if return_dict:
                    return {"hidden_states": memory}
                return memory

            # Extract the mean or last time step for prediction
            if self.config.pooling_method == "mean":
                encoder_output = tf.keras.layers.GlobalAveragePooling1D()(memory)
            elif self.config.pooling_method == "last":
                encoder_output = memory[:, -1]
            else:
                raise ValueError(f"Pooling method should be mean or last, while received {self.config.poolint_method}")

            for layer in self.dense_layers:
                encoder_output = layer(encoder_output)
            outputs = self.projection(encoder_output)
            outputs = self.reshape(outputs)

            if return_dict:
                return {"predictions": outputs}
            return outputs

        except Exception as e:
            logger.error(f"Error in forward pass: {str(e)}")
            raise



"""Time related feature: date, holiday, event"""

import logging
from typing import List, Optional

import pandas as pd

from .registry import FeatureRegistry

logger = logging.getLogger(__name__)


def add_datetime_feature():
    return



import hashlib
import json
import os
from typing import Any, Dict, List, Optional, Set, Union


class FeatureRegistry:
    def __init__(self) -> None:
        self.columns = []

    def register(self, cols):
        self.columns.extend(cols if isinstance(cols, list) else [cols])

    def get_features(self):
        return self.columns

    def hash_features(self):
        return hashlib.md5(",".join(sorted(self.columns)).encode()).hexdigest()

    def save(self, filepath: str = "feature_columns.json") -> bool:
        with open(filepath, "w") as f:
            json.dump({"features": self.columns}, f, indent=2)

    def __repr__(self) -> str:
        return f"FeatureRegistry('default') with {len(self._feature_names)} features"



"""Generate the example data script
- https://github.com/keras-team/keras/blob/v3.3.3/keras/src/utils/file_utils.py#L130-L327
"""

import logging
import random
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from tfts.constants import TFTS_ASSETS_CACHE

logger = logging.getLogger(__name__)


air_passenger_url = (
    "https://raw.githubusercontent.com/AileenNielsen/TimeSeriesAnalysisWithPython/master/data/AirPassengers.csv"
)


def get_data(
    name: str = "sine",
    train_length: int = 24,
    predict_sequence_length: int = 8,
    test_size: float = 0.1,
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[Tuple[np.ndarray, np.ndarray]], None]:
    assert (test_size >= 0) & (test_size <= 1), "test_size is the ratio of test dataset"
    if name == "sine":
        return get_sine(train_length, predict_sequence_length, test_size=test_size)

    elif name == "airpassengers":
        return get_air_passengers(train_length, predict_sequence_length, test_size=test_size)

    else:
        raise ValueError(f"unsupported data of {name} yet, try 'sine', 'airpassengers'")


def get_sine(
    train_sequence_length: int = 24, predict_sequence_length: int = 8, test_size: float = 0.2, n_examples: int = 100
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[Tuple[np.ndarray, np.ndarray]]]:
    """
    Generate synthetic sine wave data.

    Parameters:
    train_sequence_length (int): Length of the training sequence.
    predict_sequence_length (int): Length of the prediction sequence.
    test_size (float): Fraction of the data to use for validation.
    n_examples (int): Number of examples to generate.

    Returns:
    (tuple): Two tuples of numpy arrays containing training and validation data.
    """
    x: List[np.ndarray] = []
    y: List[np.ndarray] = []
    for _ in range(n_examples):
        rand = random.random() * 2 * np.pi
        sig1 = np.sin(np.linspace(rand, 3.0 * np.pi + rand, train_sequence_length + predict_sequence_length))
        sig2 = np.cos(np.linspace(rand, 3.0 * np.pi + rand, train_sequence_length + predict_sequence_length))

        x1 = sig1[:train_sequence_length]
        y1 = sig1[train_sequence_length:]
        x2 = sig2[:train_sequence_length]
        y2 = sig2[train_sequence_length:]

        x_ = np.array([x1, x2])
        y_ = np.array([y1, y2])

        x.append(x_.T)
        y.append(y_.T)

    x_array = np.array(x)[:, :, 0:1]
    y_array = np.array(y)[:, :, 0:1]
    logging.info("Load sine data", x_array.shape, y_array.shape)

    if test_size > 0:
        slice = int(n_examples * (1 - test_size))
        x_train = x_array[:slice]
        y_train = y_array[:slice]
        x_valid = x_array[slice:]
        y_valid = y_array[slice:]
        return (x_train, y_train), (x_valid, y_valid)
    return x_array, y_array


def get_air_passengers(train_sequence_length: int = 24, predict_sequence_length: int = 8, test_size: float = 0.2):
    """
    A function that loads and preprocesses the air passenger data.

    Args:
        train_sequence_length (int): The length of each input sequence.
        predict_sequence_length (int): The length of each output sequence.
        test_size (float): The fraction of the data to use for validation.

    Returns:
        Tuple of training and validation data, each containing inputs and outputs.

    """
    # air_passenger_url = "../examples/data/international-airline-passengers.csv"
    df = pd.read_csv(air_passenger_url, parse_dates=None, date_parser=None, nrows=144)
    v = df.iloc[:, 1:2].values
    v = (v - np.max(v)) / (np.max(v) - np.min(v))  # MinMaxScaler

    x: List[np.ndarray] = []
    y: List[np.ndarray] = []
    for seq in range(1, train_sequence_length + 1):
        x_roll = np.roll(v, seq, axis=0)
        x.append(x_roll)
    x_array = np.stack(x, axis=1)
    x_array = x_array[train_sequence_length:-predict_sequence_length, ::-1, :]

    for seq in range(predict_sequence_length):
        y_roll = np.roll(v, -seq)
        y.append(y_roll)
    y_array = np.stack(y, axis=1)
    y_array = y_array[train_sequence_length:-predict_sequence_length]
    logging.info("Load air passenger data", x_array.shape, y_array.shape)

    if test_size > 0:
        slice = int(len(x_array) * (1 - test_size))
        x_train = x_array[:slice]
        y_train = y_array[:slice]
        x_valid = x_array[slice:]
        y_valid = y_array[slice:]
        return (x_train, y_train), (x_valid, y_valid)
    return x_array, y_array


def get_stock_data(ticker: str = "NVDA", start_date="2023-09-01", end_date="2024-03-15") -> pd.DataFrame:
    """
    Retrieve historical stock data for a given ticker symbol.
    """
    # Download data
    import yfinance as yf

    try:
        logger.info(f"Retrieving data for {ticker} from {start_date} to {end_date}")

        data = yf.download(ticker, start=start_date, end=end_date, progress=False)

        if data.empty:
            logger.warning(f"No data returned for ticker {ticker}")
            raise ValueError(f"No data available for ticker: {ticker}")

        logger.info(f"Successfully retrieved {len(data)} records for {ticker}")
        return data

    except Exception as e:
        raise ValueError(f"Error retrieving stock data: {str(e)}")



"""Auto regression feature: lag and roll"""

import logging

from .registry import FeatureRegistry

logger = logging.getLogger(__name__)


def add_transform_feature():
    # like log1p, sqrt
    return


def add_lag_feature():
    return


def add_roll_feature():
    return



import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Type

logger = logging.getLogger(__name__)


class Pipeline(object):
    _load_processor = False

    def __init__(self, task: str, model, processor: Optional):
        self.task = task

    def __call__(self):
        pass

    def forward(self):
        pass

    def predict(self):
        pass

    def get_iterator(self):
        pass

    def postprocess(self):
        pass



"""AutoModel to choose different models"""

import collections
import importlib
import json
import logging
import os
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow as tf

from tfts.models.base import BaseConfig, BaseModel
from tfts.tasks.auto_task import AnomalyHead, ClassificationHead

from ..constants import CONFIG_NAME, TF2_WEIGHTS_INDEX_NAME, TF2_WEIGHTS_NAME, TF_WEIGHTS_NAME
from .auto_config import AutoConfig

logger = logging.getLogger(__name__)


MODEL_MAPPING_NAMES = collections.OrderedDict(
    [
        ("seq2seq", "Seq2seq"),
        ("rnn", "RNN"),
        ("wavenet", "WaveNet"),
        ("tcn", "TCN"),
        ("transformer", "Transformer"),
        ("bert", "Bert"),
        ("informer", "Informer"),
        ("autoformer", "AutoFormer"),
        ("tft", "TFTransformer"),
        ("unet", "Unet"),
        ("nbeats", "NBeats"),
        ("dlinear", "DLinear"),
        ("rwkv", "RWKV"),
    ]
)


class AutoModel(BaseModel):
    """tfts auto model
    input tensor: [batch_size, sequence_length, num_features]
    output tensor: [batch_size, predict_sequence_length, num_labels]
    """

    def __init__(self, model, config):
        super().__init__(config=config)
        self.model = model
        self.config = config

    def __call__(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """automodel callable

        Parameters
        ----------
        x : tf.data.Dataset, np.array
            model inputs
        return_dict: bool
            if return output a dict

        Returns
        -------
        tf.Tensor
            model output
        """
        if isinstance(x, (list, tuple)):
            if len(x[0].shape) != 3:
                raise ValueError(
                    f"Expected input dimension is 3 (batch_size, train_sequence_length, num_features), "
                    f"but got {len(x[0].shape)}"
                )
        if isinstance(self.model, BaseModel):
            return self.model(x, output_hidden_states=output_hidden_states, return_dict=return_dict)
        else:  # after build_model, the model will become tf.keras.Model
            return self.model(x)

    @classmethod
    def from_config(cls, config, predict_sequence_length: int = 1):
        model_name = config.model_type
        class_name = MODEL_MAPPING_NAMES[model_name]
        module = importlib.import_module(f".{model_name}", "tfts.models")
        model = getattr(module, class_name)(config=config, predict_sequence_length=predict_sequence_length)
        return cls(model, config)

    @classmethod
    def from_pretrained(cls, weights_dir: Union[str, os.PathLike], predict_sequence_length: int = 1):
        config_path = os.path.join(weights_dir, "config.json")
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"Config file not found at {config_path}")
        try:
            with open(config_path, "r") as f:
                config_dict = json.load(f)
        except Exception as e:
            raise OSError(f"Error loading config file from {config_path}. Original error: {e}")

        try:
            model_type = config_dict.get("model_type")
            if model_type is None:
                raise ValueError("Missing `model_type` in config.")

            # Dynamically get the correct Config subclass
            config = AutoConfig.for_model(model_type)
            config.update(config_dict)  # update with the saved values

            # Build model and load weights
            model = cls.from_config(config, predict_sequence_length=predict_sequence_length)
            inputs = None
            if isinstance(config.input_shape, dict):
                inputs = {k: tf.keras.layers.Input(shape=v, name=k) for k, v in config.input_shape.items()}
            elif isinstance(config.input_shape, list):
                inputs = [
                    tf.keras.layers.Input(shape=shape, name=f"input_{i}") for i, shape in enumerate(config.input_shape)
                ]
            else:
                inputs = tf.keras.layers.Input(shape=config.input_shape, name="input")

            model.build_model(inputs)
            model.model.load_weights(os.path.join(weights_dir, TF2_WEIGHTS_NAME))
            return cls(model, config)
        except Exception as e:
            raise OSError(
                f"Error loading model weights from {weights_dir}. "
                f"Ensure weights were saved using model.save_weights(...). Original error: {e}"
            )


class AutoModelForPrediction(AutoModel):
    """tfts model for prediction"""

    def __call__(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):

        model_output = self.model(x, output_hidden_states=output_hidden_states, return_dict=return_dict)

        if self.config.skip_connect_circle:
            x_mean = x[:, -self.predict_sequence_length :, 0:1]
            model_output = model_output + x_mean
        elif self.config.skip_connect_mean:
            x_mean = tf.tile(tf.reduce_mean(x[..., 0:1], axis=1, keepdims=True), [1, self.predict_sequence_length, 1])
            model_output = model_output + x_mean
        return model_output


class AutoModelForClassification(BaseModel):
    """tfts model for classification"""

    def __init__(self, model, config):
        super(AutoModelForClassification, self).__init__()
        self.model = model
        self.config = config
        self.head = ClassificationHead(num_labels=config.num_labels)
        self.keras_model: Optional[tf.keras.Model] = None

    def __call__(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        output_hidden_states: Optional[bool] = True,
        return_dict: Optional[bool] = None,
        **kwargs,
    ):
        if self.keras_model is not None:
            logits = self.keras_model(x)
        else:
            model_output = self.model(x, output_hidden_states=output_hidden_states, return_dict=return_dict)
            logits = self.head(model_output)
        return logits

    @classmethod
    def from_config(cls, config, num_labels: int = 1):
        config.num_labels = num_labels
        model_name = config.model_type
        class_name = MODEL_MAPPING_NAMES[model_name]
        module = importlib.import_module(f".{model_name}", "tfts.models")
        model = getattr(module, class_name)(config=config)
        return cls(model, config)

    def build_model(self, inputs: tf.keras.layers.Input):
        model_output = self.model(inputs)
        logits = self.head(model_output)  # Apply the head layer to the output
        self.keras_model = tf.keras.Model(inputs, logits)  # Create a complete model
        return self.keras_model


class AutoModelForAnomaly(BaseModel):
    """tfts model for anomaly detection"""

    def __init__(self, model, config):
        super().__init__(config=config)
        self.model = model
        self.config = config
        self.head = AnomalyHead(config.train_sequence_length)

    def detect(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        labels=None,
    ):
        model_output = self.model(x)
        dist = self.head(model_output, labels)
        return dist

    @classmethod
    def from_pretrained(cls, weights_dir: Union[str, os.PathLike]):
        model_path = os.path.join(weights_dir, "model.h5")
        model = tf.keras.models.load_model(model_path)
        logger.info(f"Load model from {weights_dir}")
        config_path = os.path.join(weights_dir, "config.json")
        if not os.path.exists(config_path):
            raise FileNotFoundError(f"Config file not found at {config_path}")

        config = BaseConfig.from_json(config_path)  # Load config from JSON
        return cls(model, config)

    @classmethod
    def from_config(cls, config):
        model_name = config.model_type
        class_name = MODEL_MAPPING_NAMES[model_name]
        module = importlib.import_module(f".{model_name}", "tfts.models")
        model = getattr(module, class_name)(config=config)
        return cls(model, config)


class AutoModelForSegmentation(BaseModel):
    """tfts model for time series segmentation"""

    def __init__(self, model, config):
        super().__init__(config=config)
        self.model = model
        self.config = config

    def __call__(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        return_dict: Optional[bool] = None,
        **kwargs,
    ):
        model_output = self.model(x, return_dict=return_dict)
        return model_output

    @classmethod
    def from_config(cls, config):
        model_name = config.model_type
        class_name = MODEL_MAPPING_NAMES[model_name]
        module = importlib.import_module(f".{model_name}", "tfts.models")
        model = getattr(module, class_name)(config=config)
        return cls(model, config)


class AutoModelForUncertainty(BaseModel):
    """tfts model for time series uncertainty prediction"""

    def __init__(self, model, config):
        super().__init__(config=config)
        self.model = model
        self.config = config

    def __call__(
        self,
        x: Union[tf.data.Dataset, Tuple[np.ndarray], Tuple[pd.DataFrame], List[np.ndarray], List[pd.DataFrame]],
        return_dict: Optional[bool] = None,
        **kwargs,
    ):
        model_output = self.model(x, return_dict=return_dict)
        return model_output

    @classmethod
    def from_config(cls, config):
        model_name = config.model_type
        class_name = MODEL_MAPPING_NAMES[model_name]
        module = importlib.import_module(f".{model_name}", "tfts.models")
        model = getattr(module, class_name)(config=config)
        return cls(model, config)



"""
`Long short term memory
<http://www.bioinf.jku.at/publications/older/2604.pdf>`_
"""

from typing import Dict, Literal, Optional

import tensorflow as tf
from tensorflow.keras.layers import GRU, LSTM, AveragePooling1D, Bidirectional, Concatenate, Dense, Reshape

from .base import BaseConfig, BaseModel


class RNNConfig(BaseConfig):
    model_type: str = "rnn"

    def __init__(
        self,
        rnn_hidden_size: int = 64,
        rnn_type: Literal["gru", "lstm"] = "gru",
        bi_direction: bool = False,
        dense_hidden_size: int = 128,
        num_stacked_layers: int = 1,
        scheduled_sampling: float = 0.0,
        use_attention: bool = False,
    ) -> None:
        """
        Initializes the configuration for the RNN model with the specified parameters.

        Args:
            rnn_hidden_size: The number of units in the RNN hidden layer.
            rnn_type: Type of RNN ('gru' or 'lstm').
            bi_direction: Whether to use bidirectional RNN.
            dense_hidden_size: The size of the dense hidden layer following the RNN.
            num_stacked_layers: The number of stacked RNN layers.
            scheduled_sampling: Scheduled sampling ratio.
            use_attention: Whether to use attention mechanism.
        """
        super().__init__()

        self.rnn_hidden_size: int = rnn_hidden_size
        self.rnn_type: Literal["gru", "lstm"] = rnn_type
        self.bi_direction: bool = bi_direction
        self.dense_hidden_size: int = dense_hidden_size
        self.num_stacked_layers: int = num_stacked_layers
        self.scheduled_sampling: float = scheduled_sampling
        self.use_attention: bool = use_attention


class RNN(BaseModel):
    """tfts RNN model"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[RNNConfig] = None):
        super().__init__()
        self.config = config or RNNConfig()
        self.predict_sequence_length = predict_sequence_length
        self.encoder = Encoder(rnn_size=self.config.rnn_hidden_size, rnn_type=self.config.rnn_type, return_state=True)

        self.dense1 = Dense(self.config.dense_hidden_size, activation="relu")
        self.dense2 = Dense(self.config.dense_hidden_size, activation="relu")
        self.project1 = Dense(predict_sequence_length, activation=None)

    def __call__(
        self, inputs, teacher=None, output_hidden_states: Optional[bool] = None, return_dict: Optional[bool] = None
    ):
        """RNN model call

        Parameters
        ----------
        inputs : Union[list, tuple, dict, tf.Tensor]
            Input data.
        teacher : tf.Tensor, optional
            Teacher signal for training, by default None.
        return_dict : bool, optional
            Whether to return a dictionary, by default None.

        Returns
        -------
        tf.Tensor
            Model output.
        """
        x, encoder_feature, _ = self._prepare_3d_inputs(inputs)
        encoder_outputs, encoder_state = self.encoder(encoder_feature)

        encoder_out = self.dense1(encoder_state)
        encoder_out = self.dense2(encoder_out)
        outputs = self.project1(encoder_out)
        outputs = Reshape((outputs.shape[1], 1))(outputs)
        return outputs


class Encoder(tf.keras.layers.Layer):
    def __init__(
        self,
        rnn_size: int,
        rnn_type: str = "gru",
        rnn_dropout: float = 0,
        num_stacked_layers: int = 1,
        bi_direction: bool = False,
        return_state: bool = True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.rnn_type = rnn_type.lower()
        if self.rnn_type not in ("gru", "lstm"):
            raise ValueError(f"Unsupported RNN type: {self.rnn_type}")

        self.rnn_size = rnn_size
        self.rnn_dropout = rnn_dropout
        self.num_stacked_layers = num_stacked_layers
        self.bi_direction = bi_direction
        self.return_state = return_state
        self.rnn_layers = []

    def build(self, input_shape):
        for i in range(self.num_stacked_layers):
            return_state = self.return_state if i == self.num_stacked_layers - 1 and not self.bi_direction else False

            rnn_class = GRU if self.rnn_type == "gru" else LSTM
            rnn = rnn_class(
                units=self.rnn_size,
                activation="tanh",
                return_sequences=True,
                return_state=return_state,
                dropout=self.rnn_dropout if self.rnn_dropout > 0 else 0.0,
            )

            if self.bi_direction:
                rnn = Bidirectional(rnn)

            self.rnn_layers.append(rnn)

        super().build(input_shape)

    def call(self, inputs: tf.Tensor):
        """RNN encoder call

        Parameters
        ----------
        inputs : tf.Tensor
            3d time series input, (batch_size, sequence_length, num_features)

        Returns
        -------
        tf.Tensor or tuple
            If return_state is False:
                output tensor: (batch_size, sequence_length, rnn_size)
            If return_state is True:
                For GRU: (output, state)
                    output: (batch_size, sequence_length, rnn_size)
                    state: (batch_size, rnn_size)
                For LSTM: (output, state_h, state_c)
                    output: (batch_size, sequence_length, rnn_size)
                    state_h: (batch_size, rnn_size)
                    state_c: (batch_size, rnn_size)
        """
        x = inputs

        for i, layer in enumerate(self.rnn_layers):
            is_last_layer = i == len(self.rnn_layers) - 1
            if is_last_layer and self.return_state:
                outputs = layer(x)

                if self.bi_direction:
                    output = outputs[0]
                    if self.rnn_type == "gru":
                        # GRU: forward_state, backward_state
                        fw_state, bw_state = outputs[1:]
                        state = tf.concat([fw_state, bw_state], axis=-1)
                        return output, state
                    else:  # LSTM
                        fw_h, fw_c, bw_h, bw_c = outputs[1:]
                        state_h = tf.concat([fw_h, bw_h], axis=-1)
                        state_c = tf.concat([fw_c, bw_c], axis=-1)
                        state = Concatenate(axis=-1)([state_h, state_c])
                        return output, state
                else:
                    if self.rnn_type == "lstm":
                        x, state_h, state_c = outputs
                        state = Concatenate(axis=-1)([state_h, state_c])
                    else:
                        x, state = outputs
                    return x, state
            else:
                x = layer(x)
        return x

    def get_config(self):
        config = {
            "rnn_type": self.rnn_type,
            "rnn_size": self.rnn_size,
            "rnn_dropout": self.rnn_dropout,
        }
        base_config = super(Encoder, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))

    def compute_output_shape(self, input_shape):
        """Compute the output shape of the encoder."""
        batch_size, seq_length, _ = input_shape

        if self.bi_direction:
            rnn_output_size = 2 * self.rnn_size
        else:
            rnn_output_size = self.rnn_size

        if not self.return_state:
            return (batch_size, seq_length, rnn_output_size)

        else:
            if self.bi_direction:
                # Bidirectional with return_state is always just the output sequence
                return (batch_size, seq_length, rnn_output_size)
            elif self.rnn_type == "gru":
                # GRU: (output, state)
                return ((batch_size, seq_length, rnn_output_size), (batch_size, rnn_output_size))
            else:  # LSTM
                # LSTM: (output, state_h, state_c)
                return ((batch_size, seq_length, rnn_output_size), (batch_size, 2 * rnn_output_size))


class RNN2(object):
    def __init__(self, predict_sequence_length=3, config=RNNConfig()) -> None:
        self.config = config
        self.predict_sequence_length = predict_sequence_length
        self.rnn = GRU(
            units=config.rnn_hidden_size, activation="tanh", return_state=True, return_sequences=True, dropout=0
        )
        self.dense1 = Dense(predict_sequence_length)
        self.dense2 = Dense(1)

    def __call__(self, inputs: tf.Tensor, teacher: Optional[tf.Tensor] = None):
        """Another RNN model

        Parameters
        ----------
        inputs : _type_
            _description_
        teacher : _type_, optional
            _description_, by default None

        Returns
        -------
        _type_
            _description_
        """
        if isinstance(inputs, (list, tuple)):
            x, encoder_features, _ = inputs
            encoder_features = tf.concat([x, encoder_features], axis=-1)
        else:  # for single variable prediction
            encoder_features = x = inputs

        encoder_shape = tf.shape(encoder_features)
        future = tf.zeros([encoder_shape[0], self.predict_sequence_length, encoder_shape[2]])
        encoder_features = tf.concat([encoder_features, future], axis=1)
        output, state = self.rnn(encoder_features)
        output = self.dense2(output)

        return output[:, -self.predict_sequence_length :]



"""Time series task head"""

from typing import Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Dense, GlobalAveragePooling1D

from .base import BaseTask


class PredictionHead(tf.keras.layers.Layer, BaseTask):
    """Prediction task head layer"""

    def __init__(self):
        super(PredictionHead, self).__init__()


class SegmentationHead(tf.keras.layers.Layer, BaseTask):
    """Segmentation task head layer"""

    def __init__(self):
        super(SegmentationHead, self).__init__()


class ClassificationHead(tf.keras.layers.Layer):
    """Classification task head layer"""

    def __init__(self, num_labels: int = 1, dense_units: Tuple[int] = (128,)):
        super(ClassificationHead, self).__init__()
        self.pooling = GlobalAveragePooling1D(data_format="channels_last")
        self.intermediate_dense_layers = []
        for unit in dense_units:
            self.intermediate_dense_layers.append(Dense(unit, activation="relu"))
        self.classifier = Dense(num_labels, activation="softmax")

    def call(self, inputs: tf.Tensor, **kwargs) -> tf.Tensor:
        """classification task head

        Parameters
        ----------
        inputs : tf.Tensor
            model backbone output as task input, (batch_size, train_sequence_length, hidden_size)

        Returns
        -------
        tf.Tensor
            logit of the classification
        """
        # => (batch_size, hidden_size)
        pooled_output = self.pooling(inputs)

        for layer in self.intermediate_dense_layers:
            pooled_output = layer(pooled_output)

        # => (batch_size, num_labels)
        logits = self.classifier(pooled_output)
        return logits


class AnomalyHead:
    """Anomaly task head layer: Reconstruct style"""

    def __init__(self, train_sequence_length: int) -> None:
        super().__init__()
        self.train_sequence_length = train_sequence_length

    def __call__(self, y_pred, y_test):
        if isinstance(y_pred, tf.Tensor):
            y_pred = y_pred.numpy()
        if isinstance(y_test, tf.Tensor):
            y_test = y_test.numpy()
        if y_pred.shape[1] == 1:
            y_pred = np.squeeze(y_pred, 1)
        errors = y_pred - y_test

        # mean / cov
        mean = sum(errors) / len(errors)
        cov = 0
        for e in errors:
            cov += np.dot((e - mean).reshape(len(e), 1), (e - mean).reshape(1, len(e)))
        cov /= len(errors)

        m_dist = [0] * self.train_sequence_length
        for e in errors:
            m_dist.append(AnomalyHead.mahala_distantce(e, mean, cov))

        return m_dist

    @staticmethod
    def mahala_distantce(x, mean, cov, epsilon=1e-8):
        cov += epsilon * np.eye(cov.shape[0])  # Zero Covariance
        d = np.dot(x - mean, np.linalg.inv(cov))
        d = np.dot(d, (x - mean).T)
        return d


class GaussianHead(tf.keras.layers.Layer):
    def __init__(self, units: int):
        self.units = units
        super().__init__()

    def build(self, input_shape: Tuple[Optional[int], ...]):
        in_channels = input_shape[2]
        self.weight1 = self.add_weight(
            name="gauss_w1", shape=(in_channels, self.units), initializer=tf.keras.initializers.GlorotNormal()
        )
        self.weight2 = self.add_weight(
            name="gauss_w2", shape=(in_channels, self.units), initializer=tf.keras.initializers.GlorotNormal()
        )
        self.bias1 = self.add_weight(name="gauss_b1", shape=(self.units,), initializer=tf.keras.initializers.Zeros())
        self.bias2 = self.add_weight(name="gauss_b2", shape=(self.units,), initializer=tf.keras.initializers.Zeros())
        super().build(input_shape)

    def call(self, x: tf.Tensor):
        """Returns mean and standard deviation tensors.

        Args:
          x (tf.Tensor): Input tensor.

        Returns:
          Tuple[tf.Tensor, tf.Tensor]: Mean and standard deviation tensors.
        """
        mu = tf.matmul(x, self.weight1) + self.bias1
        sig = tf.matmul(x, self.weight2) + self.bias2
        sig_pos = tf.math.log1p(tf.math.exp(sig)) + 1e-7
        return mu, sig_pos

    def get_config(self):
        """Returns the configuration of the layer."""
        config = {"units": self.units}
        base_config = super().get_config()
        return {**base_config, **config}



"""
`WaveNet: A Generative Model for Raw Audio
<https://arxiv.org/abs/1609.03499>`_
"""

import logging
from typing import Dict, List, Optional

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Concatenate, Dense, Lambda, ReLU

from tfts.layers.cnn_layer import ConvTemp
from tfts.layers.dense_layer import DenseTemp

from .base import BaseConfig, BaseModel

logger = logging.getLogger(__name__)


class WaveNetConfig(BaseConfig):
    model_type: str = "wavenet"

    def __init__(
        self,
        dilation_rates: List[int] = None,
        kernel_sizes: List[int] = None,
        filters: int = 128,
        dense_hidden_size: int = 64,
        scheduled_sampling: float = 1.0,
        use_attention: bool = False,
        attention_size: int = 64,
        num_attention_heads: int = 2,
        attention_probs_dropout_prob: float = 0.0,
        **kwargs,
    ) -> None:
        """
        Initializes the configuration for the WaveNet model with the specified parameters.

        Args:
            dilation_rates: List of dilation rates for the convolutional layers.
            kernel_sizes: List of kernel sizes for the convolutional layers.
            filters: The number of filters in the convolutional layers.
            dense_hidden_size: The size of the dense hidden layer following the convolutional layers.
            scheduled_sampling: Scheduled sampling ratio. 0 means teacher forcing, 1 means use last prediction
            use_attention: Whether to use attention mechanism in the model.
            attention_size: The size of the attention mechanism.
            num_attention_heads: The number of attention heads.
            attention_probs_dropout_prob: Dropout probability for attention probabilities.
        """
        super(WaveNetConfig, self).__init__()

        self.dilation_rates: List[int] = dilation_rates or [2**i for i in range(4)]
        self.kernel_sizes: List[int] = kernel_sizes or [2] * 4
        self.filters: int = filters
        self.dense_hidden_size: int = dense_hidden_size
        self.scheduled_sampling: float = scheduled_sampling
        self.use_attention: bool = use_attention
        self.attention_size: int = attention_size
        self.num_attention_heads: int = num_attention_heads
        self.attention_probs_dropout_prob: float = attention_probs_dropout_prob


class WaveNet(BaseModel):
    """WaveNet model for time series"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[WaveNetConfig] = None) -> None:
        """
        Initializes the WaveNet model.

        Args:
            predict_sequence_length: Length of the prediction sequence.
            config: Configuration object containing model parameters.
        """
        super(WaveNet, self).__init__()
        self.config = config or WaveNetConfig()
        self.predict_sequence_length = predict_sequence_length
        self.encoder = Encoder(
            kernel_sizes=self.config.kernel_sizes,
            dilation_rates=self.config.dilation_rates,
            filters=self.config.filters,
            dense_hidden_size=self.config.dense_hidden_size,
        )
        self.decoder = DecoderV1(
            filters=self.config.filters,
            dilation_rates=self.config.dilation_rates,
            dense_hidden_size=self.config.dense_hidden_size,
            predict_sequence_length=self.predict_sequence_length,
        )

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """
        Forward pass for the WaveNet model.

        Args:
            inputs: Input tensor for the model.
            teacher: Teacher tensor used for scheduled sampling.
            output_hidden_states: Flag to output the hidden statues
            return_dict: Flag to control the return type.

        Returns:
            Tensor containing the model output.
        """
        x, encoder_feature, decoder_feature = self._prepare_3d_inputs(inputs, ignore_decoder_inputs=False)
        encoder_state, encoder_outputs = self.encoder(encoder_feature)
        decoder_outputs = self.decoder(
            decoder_features=decoder_feature,
            decoder_init_input=x[:, -1],
            teacher=teacher,
            encoder_outputs=encoder_outputs,
        )

        return decoder_outputs


class Encoder(object):
    """Encoder block for the WaveNet model."""

    def __init__(
        self, kernel_sizes: List[int], filters: int, dilation_rates: List[int], dense_hidden_size: int
    ) -> None:
        """
        Initializes the encoder block.

        Args:
            kernel_sizes: List of kernel sizes for convolutional layers.
            filters: Number of filters for convolutional layers.
            dilation_rates: Dilation rates for the convolutions.
            dense_hidden_size: Hidden size for the dense layers.
        """
        self.filters = filters
        self.conv_times = []
        for i, (kernel_size, dilation) in enumerate(zip(kernel_sizes, dilation_rates)):
            self.conv_times.append(
                ConvTemp(filters=2 * filters, kernel_size=kernel_size, causal=True, dilation_rate=dilation)
            )
        self.dense_time1 = DenseTemp(hidden_size=filters, activation="tanh", name="encoder_dense_time1")
        self.dense_time2 = DenseTemp(hidden_size=filters + filters, name="encoder_dense_time2")
        self.dense_time3 = DenseTemp(hidden_size=dense_hidden_size, activation="relu", name="encoder_dense_time3")
        self.dense_time4 = DenseTemp(hidden_size=1, name="encoder_dense_time_4")

    def __call__(self, x: tf.Tensor):
        inputs = self.dense_time1(inputs=x)

        skip_outputs = []
        conv_inputs = [inputs]
        for conv_time in self.conv_times:
            dilated_conv = conv_time(inputs)
            split_layer = Lambda(lambda x: tf.split(x, 2, axis=2))
            conv_filter, conv_gate = split_layer(dilated_conv)
            dilated_conv = Lambda(lambda x: tf.nn.tanh(x[0]) * tf.nn.sigmoid(x[1]))([conv_filter, conv_gate])
            outputs = self.dense_time2(inputs=dilated_conv)
            split_layer2 = Lambda(lambda x: tf.split(x, [self.filters, self.filters], axis=2))
            skips, residuals = split_layer2(outputs)
            inputs += residuals
            conv_inputs.append(inputs)  # batch_size * time_sequence_length * filters
            skip_outputs.append(skips)

        concat_layer = Concatenate(axis=2)
        concatenated = concat_layer(skip_outputs)
        relu_layer = ReLU()
        skip_outputs = relu_layer(concatenated)
        # skip_outputs = tf.nn.relu(tf.concat(skip_outputs, axis=2))
        h = self.dense_time3(skip_outputs)
        # [batch_size, time_sequence_length, filters] * time_sequence_length
        y_hat = self.dense_time4(h)
        return y_hat, conv_inputs[:-1]


class DecoderV1(object):
    """Decoder block for WaveNet V1."""

    def __init__(
        self, filters: int, dilation_rates: List[int], dense_hidden_size: int, predict_sequence_length: int = 24
    ) -> None:
        """
        Initializes the decoder block.

        Args:
            filters: Number of filters for convolutional layers.
            dilation_rates: Dilation rates for convolutions.
            dense_hidden_size: Size of the dense hidden layer.
            predict_sequence_length: Length of the predicted sequence.
        """
        self.filters: int = filters
        self.predict_sequence_length = predict_sequence_length
        self.dilation_rates = dilation_rates
        self.dense1 = Dense(filters, activation="tanh")
        self.dense2 = Dense(2 * filters, use_bias=True)
        self.dense3 = Dense(2 * filters, use_bias=False)
        self.dense4 = Dense(2 * filters)
        self.dense5 = Dense(dense_hidden_size, activation="relu")
        self.dense6 = Dense(1)

    def __call__(
        self,
        decoder_features,
        decoder_init_input,
        encoder_outputs,
        teacher: Optional[tf.Tensor] = None,
        scheduled_sampling: float = 0.0,
        training: Optional[bool] = None,
        **kwargs: Dict,
    ):
        """
        Forward pass for the decoder block.

        Args:
            decoder_features: Tensor containing decoder features.
            decoder_init_input: Initial input for the decoder.
            encoder_outputs: List of encoder outputs.
            teacher: Optional tensor for teacher forcing.
            scheduled_sampling: Probability of using teacher forcing.
            training: Whether the model is in training mode.

        Returns:
            Decoder output tensor.
        """
        decoder_outputs = []
        prev_output = decoder_init_input  # the initial input for decoder

        for i in range(self.predict_sequence_length):
            if training:
                p = np.random.uniform(low=0, high=1, size=1)[0]
                if teacher is not None and p > scheduled_sampling:
                    this_input = teacher[:, i : i + 1]
                else:
                    this_input = prev_output
            else:
                this_input = prev_output

            if decoder_features is not None:
                # this_input = tf.concat([this_input, decoder_features[:, i]], axis=-1)
                this_input = Concatenate(axis=-1)([this_input, decoder_features[:, i]])

            x = self.dense1(this_input)
            skip_outputs = []

            for i, dilation in enumerate(self.dilation_rates):
                state = encoder_outputs[i][:, -dilation, :]
                # use 2 dense layer to calculate a kernel=2 convolution
                dilated_conv = self.dense2(state) + self.dense3(x)
                # conv_filter, conv_gate = tf.split(dilated_conv, 2, axis=1)
                split_layer = Lambda(lambda x: tf.split(x, 2, axis=1))
                conv_filter, conv_gate = split_layer(dilated_conv)
                # dilated_conv = tf.nn.tanh(conv_filter) * tf.nn.sigmoid(conv_gate)
                dilated_conv = Lambda(lambda x: tf.nn.tanh(x[0]) * tf.nn.sigmoid(x[1]))([conv_filter, conv_gate])

                out = self.dense4(dilated_conv)
                # skip, residual = tf.split(out, 2, axis=1)
                split_layer = Lambda(lambda x: tf.split(x, [self.filters, self.filters], axis=1))
                skips, residuals = split_layer(out)
                x += residuals
                # encoder_outputs[i] = tf.concat([encoder_outputs[i], tf.expand_dims(x, 1)], axis=1)
                expand = Lambda(lambda t: tf.expand_dims(t, axis=1))
                encoder_outputs[i] = Concatenate(1)([encoder_outputs[i], expand(x)])
                skip_outputs.append(skips)

            # skip_outputs = tf.nn.relu(tf.concat(skip_outputs, axis=1))
            concat_layer = Concatenate(axis=1)
            concatenated = concat_layer(skip_outputs)
            relu_layer = ReLU()
            skip_outputs = relu_layer(concatenated)
            skip_outputs = self.dense5(skip_outputs)
            this_output = self.dense6(skip_outputs)
            decoder_outputs.append(this_output)

        # decoder_outputs = tf.concat(decoder_outputs, axis=1)
        decoder_outputs = Concatenate(1)(decoder_outputs)
        expand = Lambda(lambda t: tf.expand_dims(t, axis=-1))
        return expand(decoder_outputs)


class DecoderV2(object):
    """Decoder need avoid future data leaks"""

    def __init__(
        self, filters: int, dilation_rates: List[int], dense_hidden_size: int, predict_sequence_length: int = 24
    ):
        self.filters = filters
        self.dilation_rates = dilation_rates
        self.predict_sequence_length = predict_sequence_length
        self.dense_1 = Dense(filters, activation="tanh", name="decoder_dense_1")
        self.dense_2 = Dense(2 * filters, name="decoder_dense_2")
        self.dense_3 = Dense(2 * filters, use_bias=False, name="decoder_dense_3")
        self.dense_4 = Dense(2 * filters, name="decoder_dense_4")
        self.dense_5 = Dense(dense_hidden_size, activation="relu", name="decoder_dense_5")
        self.dense_6 = Dense(1, name="decoder_dense_6")

    def __call__(
        self,
        decoder_features: tf.Tensor,
        decoder_init_input: tf.Tensor,
        encoder_states: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
    ):
        """
        Forward pass for the decoder block v2.

        Args:
            decoder_features: Tensor containing decoder features.
            decoder_init_input: Initial input for the decoder.
            encoder_states: List of encoder outputs.
            teacher: Optional tensor for teacher forcing.

        Returns:
            Decoder output tensor.
        """

        def cond_fn(time, prev_output, decoder_output_ta):
            return time < self.predict_sequence_length

        def body(time, prev_output, decoder_output_ta):
            if time == 0 or teacher is None:
                current_input = prev_output
            else:
                current_input = teacher[:, time - 1, :]

            if decoder_features is not None:
                current_feature = decoder_features[:, time, :]
                current_input = tf.concat([current_input, current_feature], axis=1)

            inputs = self.dense_1(current_input)

            skip_outputs = []
            for i, dilation in enumerate(self.dilation_rates):
                state = encoder_states[i][:, -dilation, :]

                dilated_conv = self.dense_2(state) + self.dense_3(inputs)
                conv_filter, conv_gate = tf.split(dilated_conv, 2, axis=1)
                dilated_conv = tf.nn.tanh(conv_filter) * tf.nn.sigmoid(conv_gate)
                outputs = self.dense_4(dilated_conv)
                skips, residuals = tf.split(outputs, [self.filters, self.filters], axis=1)
                inputs += residuals
                encoder_states[i] = tf.concat([encoder_states[i], tf.expand_dims(inputs, 1)], axis=1)
                skip_outputs.append(skips)

            skip_outputs = tf.nn.relu(tf.concat(skip_outputs, axis=1))
            h = self.dense_5(skip_outputs)
            y_hat = self.dense_6(h)
            decoder_output_ta = decoder_output_ta.write(time, y_hat)
            return time + 1, y_hat, decoder_output_ta

        loop_init = [
            tf.constant(0, dtype=tf.int32),
            decoder_init_input,
            tf.TensorArray(dtype=tf.float32, size=self.predict_sequence_length),
        ]
        _, _, decoder_outputs_ta = tf.while_loop(cond=cond_fn, body=body, loop_vars=loop_init)
        decoder_outputs = decoder_outputs_ta.stack()
        decoder_outputs = tf.transpose(decoder_outputs, [1, 0, 2])
        return decoder_outputs



"""Layer for :py:class:`~tfts.models.unet`"""

from typing import Dict, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Activation, Add, BatchNormalization, Conv1D, Dense, GlobalAveragePooling1D, Multiply


class ConvbrLayer(tf.keras.layers.Layer):
    def __init__(self, units: int, kernel_size: int, strides: int, dilation: int):
        super(ConvbrLayer, self).__init__()
        self.units = units
        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation = dilation

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.conv1 = Conv1D(
            self.units, kernel_size=self.kernel_size, strides=self.strides, dilation_rate=self.dilation, padding="same"
        )
        self.bn = BatchNormalization()
        self.relu = Activation("relu")
        super(ConvbrLayer, self).build(input_shape)

    def call(self, x):
        """_summary_

        Parameters
        ----------
        x : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        x = self.conv1(x)
        x = self.bn(x)
        x = self.relu(x)
        return x

    def get_config(self):
        config = {"units": self.units, "kernel_size": self.kernel_size}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class SeBlock(tf.keras.layers.Layer):
    """Squeeze-and-Excitation Networks"""

    def __init__(self, units):
        super(SeBlock, self).__init__()
        self.units = units

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.pool = GlobalAveragePooling1D()
        self.fc1 = Dense(self.units // 8, activation="relu")
        self.fc2 = Dense(self.units, activation="sigmoid")
        super(SeBlock, self).build(input_shape)

    def call(self, x):
        """_summary_

        Parameters
        ----------
        x : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        input = x
        x = self.pool(x)
        x = self.fc1(x)
        x = self.fc2(x)
        x_out = Multiply()([input, x])
        return x_out

    def get_config(self):
        config = {"units": self.units}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


class ReBlock(tf.keras.layers.Layer):
    def __init__(self, units, kernel_size, strides, dilation, use_se):
        super(ReBlock, self).__init__()
        self.units = units
        self.kernel_size = kernel_size
        self.strides = strides
        self.dilation = dilation
        self.use_se = use_se

    def build(self, input_shape: Tuple[Optional[int], ...]):
        super().build(input_shape)
        self.conv_br1 = ConvbrLayer(self.units, self.kernel_size, self.strides, self.dilation)
        self.conv_br2 = ConvbrLayer(self.units, self.kernel_size, self.strides, self.dilation)
        if self.use_se:
            self.se_block = SeBlock(units=self.units)

    def call(self, x):
        """_summary_

        Parameters
        ----------
        x : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        x_re = self.conv_br1(x)
        x_re = self.conv_br2(x_re)
        if self.use_se:
            x_re = self.se_block(x_re)
            x_re = Add()([x, x_re])
        return x_re

    def get_config(self):
        config = {"units": self.units}
        base_config = super().get_config()
        return dict(list(base_config.items()) + list(config.items()))


def conv_br(x, units, kernel_size, strides=1, dilation=1):
    # a function is easier to reuse
    convbr = ConvbrLayer(units=units, kernel_size=kernel_size, strides=strides, dilation=dilation)
    out = convbr(x)
    return out


def se_block(x, units):
    seblock = SeBlock(units)
    out = seblock(x)
    return out


def re_block(x, units, kernel_size, strides=1, dilation=1, use_se=True):
    reblock = ReBlock(units, kernel_size, strides, dilation, use_se=use_se)
    out = reblock(x)
    return out



"""Base class for config and model"""

from abc import ABC, abstractmethod
import collections
import json
import logging
import os
from typing import Any, Dict, Optional, Union

import tensorflow as tf
from tensorflow.keras.layers import Concatenate, Lambda

from ..constants import CONFIG_NAME, TF2_WEIGHTS_INDEX_NAME, TF2_WEIGHTS_NAME, TF_WEIGHTS_NAME

logger = logging.getLogger(__name__)


class BaseModel(ABC):
    """Bert model for time series forecasting.

    This model implements a transformer-based architecture (BERT) adapted for time series data.
    It processes time series inputs through a transformer encoder and produces predictions
    for future time steps.

    Parameters
    ----------
    predict_sequence_length : int, optional
        Number of future time steps to predict, by default 1
    config : BertConfig, optional
        Configuration parameters for the model, by default None
    """

    def __init__(self, predict_sequence_length: int = 1, config: Optional["BaseConfig"] = None):
        self.config = config
        self.predict_sequence_length = predict_sequence_length
        self.model = None  # Model should be defined later (may not be directly used in all subclasses)

    def build_model(self, inputs: tf.keras.layers.Input) -> tf.keras.Model:
        if hasattr(self, "config"):
            if isinstance(inputs, dict):
                self.config.input_shape = {k: tuple(v.shape[1:]) for k, v in inputs.items()}
            elif isinstance(inputs, (list, tuple)):
                self.config.input_shape = [tuple(v.shape[1:]) for v in inputs]
            else:
                self.config.input_shape = tuple(inputs.shape[1:])

        # only accept the inputs parameters after built
        outputs = self.model(inputs)
        # to handles the Keras symbolic tensors for tf2.3.1, use []
        self.model = tf.keras.Model(inputs, outputs)
        return self.model

    def to_model(self):
        inputs = tf.keras.Input(shape=(self.config.input_shape))
        return self.build_model(inputs)

    def load_pretrained_weights(self, weights_dir: str):
        if not os.path.exists(weights_dir):
            raise FileNotFoundError(f"Weights file not found at {weights_dir}")
        self.model = tf.keras.models.load_model(weights_dir)
        # self.model = model.load_weights(os.path.join(weights_dir, "weights.h5"))

    def _prepare_3d_inputs(self, inputs, ignore_decoder_inputs=True):
        """
        Prepares 3D inputs for model processing by extracting and formatting features from various input types.

        Args:
            inputs: Input data that can be a tuple/list, dictionary, or tensor.
                - If tuple/list: Expected to be [x, encoder_feature, decoder_feature]
                - If dictionary: Expected to have keys "x", "encoder_feature", and "decoder_feature"
                - If tensor: Used directly as both x and encoder_feature

        Returns:
            tuple: (x, encoder_feature, decoder_feature) properly formatted for model processing
        """

        decoder_feature = None
        if isinstance(inputs, (list, tuple)):
            x, encoder_feature, decoder_feature = inputs
            encoder_feature = Concatenate(axis=-1)([x, encoder_feature])
        elif isinstance(inputs, dict):
            x = inputs["x"]
            encoder_feature = inputs["encoder_feature"]
            encoder_feature = Concatenate(axis=-1)([x, encoder_feature])
            if "decoder_feature" in inputs:
                decoder_feature = inputs["decoder_feature"]
        else:
            encoder_feature = x = inputs
            if not ignore_decoder_inputs:
                decoder_feature = Lambda(
                    lambda encoder_feature: tf.cast(
                        tf.tile(
                            tf.reshape(tf.range(self.predict_sequence_length), (1, self.predict_sequence_length, 1)),
                            (tf.shape(encoder_feature)[0], 1, 1),
                        ),
                        tf.float32,
                    )
                )(encoder_feature)
        return x, encoder_feature, decoder_feature

    def save_pretrained(
        self,
        save_directory: Union[str, os.PathLike],
        max_shard_size: Union[int, str] = "8GB",
        safe_serialization: bool = False,
    ):
        if os.path.isfile(save_directory):
            logger.error(f"Provided path ({save_directory}) should be a directory, not a file")
            return

        os.makedirs(save_directory, exist_ok=True)
        self.config.architectures = [self.__class__.__name__[2:]]
        self.config.save_pretrained(save_directory)

        weights_file = os.path.join(save_directory, TF2_WEIGHTS_NAME)  # Or the appropriate extension

        try:
            self.model.save_weights(weights_file)
            logging.info(f"Model weights successfully saved in {weights_file}")
        except Exception as e:
            logging.error(f"Failed to save model weights to {weights_file}: {e}")
            return

    def save_weights(self, weights_path: str):
        if weights_path.endswith(".h5"):
            # User passed a full filepath
            weights_file = weights_path
            config_file = weights_path.replace(".h5", ".config.json")
            os.makedirs(os.path.dirname(weights_file), exist_ok=True)
        else:
            # User passed a directory
            os.makedirs(weights_path, exist_ok=True)
            weights_file = os.path.join(weights_path, TF2_WEIGHTS_NAME)
            config_file = os.path.join(weights_path, CONFIG_NAME)

        self.model.save_weights(weights_file)
        self.config.to_json(config_file)
        logger.info(f"Model weights successfully saved in {weights_file}")

    def save_model(self, weights_dir: str):
        self.model.save(weights_dir)
        logger.info(f"Protobuf model successfully saved in {weights_dir}")

    def summary(self):
        if hasattr(self, "model") and self.model is not None:
            self.model.summary()
        else:
            raise RuntimeError("Model has not been built yet. Please build the model first.")


class BaseConfig(ABC):
    """Base class for tfts config."""

    attribute_map: Dict[str, str] = {}
    model_type: str

    def __init__(self, **kwargs):
        self.update(kwargs)

    def __setattr__(self, key: str, value):
        mapped_key = self.attribute_map.get(key, key)
        super().__setattr__(mapped_key, value)

    def __getattribute__(self, key: str):
        if key != "attribute_map" and key in super().__getattribute__("attribute_map"):
            key = super().__getattribute__("attribute_map")[key]
        return super().__getattribute__(key)

    def update(self, config_dict: Dict[str, Any]):
        for key, value in config_dict.items():
            try:
                setattr(self, key, value)
            except AttributeError as err:
                logger.error(f"Can't set {key} with value {value} for {self}")
                raise err

    def to_dict(self):
        instance_attributes = {key: getattr(self, key) for key in self.__dict__ if not key.startswith("_")}

        if hasattr(self, "model_type"):
            instance_attributes["model_type"] = self.model_type
        return instance_attributes

    def to_json(self, json_file: Union[str, os.PathLike]):
        config_dict = self.to_dict()
        with open(json_file, "w", encoding="utf-8") as file:
            json.dump(config_dict, file, indent=4)

    @classmethod
    def from_dict(cls, config_dict: Dict[str, Any]):
        return cls(**config_dict)

    @classmethod
    def from_json(cls, json_file: Union[str, os.PathLike]):
        with open(json_file, "r", encoding="utf-8") as reader:
            text = reader.read()
        config_dict = json.loads(text)
        return cls(**config_dict)

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: Union[str, os.PathLike],
        cache_dir: Optional[Union[str, os.PathLike]] = None,
        force_download: bool = False,
    ):
        with open(pretrained_model_name_or_path, "r") as f:
            config_dict = json.load(f)
        return cls.from_dict(config_dict)

    def save_pretrained(self, save_directory: Union[str, os.PathLike]):
        os.makedirs(save_directory, exist_ok=True)
        output_config_file = os.path.join(save_directory, CONFIG_NAME)

        try:
            with open(output_config_file, "w") as f:
                json.dump(self.to_dict(), f, indent=4)
            logger.info(f"Model config successfully saved in {output_config_file}")
        except Exception as e:
            logger.warning(f"Error saving model config to {output_config_file}: {e}")

    def __str__(self):
        """Convert config to string representation in dictionary format"""
        return str({k: v for k, v in self.__dict__.items() if not k.startswith("_")})


def flatten_dict(nested, sep="/"):
    """Flatten dictionary and concatenate nested keys with separator."""

    def rec(nest, prefix, into):
        for k, v in nest.items():
            if sep in k:
                raise ValueError(f"separator '{sep}' not allowed to be in key '{k}'")
            if isinstance(v, collections.Mapping):
                rec(v, prefix + k + sep, into)
            else:
                into[prefix + k] = v

    flat = {}
    rec(nested, "", flat)
    return flat



"""
`Autoformer: Decomposition Transformers with Auto-Correlation for Long-Term Series Forecasting
<https://arxiv.org/abs/2106.13008>`_
"""

from typing import Any, Callable, Dict, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Conv1D, Dense, Dropout, LayerNormalization, ReLU

from tfts.layers.attention_layer import Attention, SelfAttention
from tfts.layers.autoformer_layer import AutoCorrelation, SeriesDecomp
from tfts.layers.dense_layer import FeedForwardNetwork
from tfts.layers.embed_layer import DataEmbedding


class AutoFormerConfig(object):
    """
    Configuration class to store the configuration of a [`AutoFormer`]
    """

    model_type: str = "autoformer"

    def __init__(
        self,
        kernel_size=25,
        hidden_size=64,
        num_layers=1,
        num_decoder_layers=None,
        num_attention_heads=4,
        ffn_intermediate_size=256,
        hidden_act="relu",
        hidden_dropout_prob=0.0,
        attention_probs_dropout_prob=0.0,
        max_position_embeddings=512,
        type_vocab_size=2,
        initializer_range=0.02,
        layer_norm_eps=1e-12,
        positional_type="absolute",
        use_cache=True,
        classifier_dropout=None,
        **kwargs,
    ):
        self.kernel_size = kernel_size
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.num_decoder_layers = num_decoder_layers if num_decoder_layers is not None else self.num_layers
        self.num_attention_heads = num_attention_heads
        self.hidden_act = hidden_act
        self.ffn_intermediate_size = ffn_intermediate_size
        self.hidden_dropout_prob = hidden_dropout_prob
        self.attention_probs_dropout_prob = attention_probs_dropout_prob
        self.max_position_embeddings = max_position_embeddings
        self.type_vocab_size = type_vocab_size
        self.initializer_range = initializer_range
        self.layer_norm_eps = layer_norm_eps
        self.positional_type = positional_type
        self.use_cache = use_cache
        self.classifier_dropout = classifier_dropout


class AutoFormer(object):
    """AutoFormer model"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[AutoFormerConfig] = None) -> None:
        self.config = config or AutoFormerConfig()
        self.predict_sequence_length = predict_sequence_length

        # self.encoder_embedding = TokenEmbedding(config['hidden_size'])
        self.series_decomp = SeriesDecomp(self.config.kernel_size)
        self.encoder = [
            EncoderLayer(
                self.config.kernel_size,
                self.config.hidden_size,
                self.config.num_attention_heads,
                self.config.attention_probs_dropout_prob,
            )
            for _ in range(self.config.num_layers)
        ]

        self.decoder = [
            DecoderLayer(
                self.config.kernel_size,
                self.config.hidden_size,
                self.config.num_attention_heads,
                self.config.attention_probs_dropout_prob,
            )
            for _ in range(self.config.num_decoder_layers)
        ]

        self.project = Conv1D(1, kernel_size=3, strides=1, padding="same", use_bias=False)
        self.project1 = Dense(predict_sequence_length, activation=None)
        self.drop1 = Dropout(0.0)
        self.dense1 = Dense(512, activation="relu")
        self.drop2 = Dropout(0.0)
        self.dense2 = Dense(1024, activation="relu")

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """Forward pass of the AutoFormer model.

        Args:
            inputs: Input tensor(s) or dictionary.
                Can be:
                - A single tensor of shape [batch_size, time_steps, features]
                - A list/tuple of (x, encoder_feature, decoder_feature)
                - A dictionary with keys 'x' and 'encoder_feature'
            teacher: Optional teacher forcing input.
            training: Whether the model is in training mode.
            output_hidden_states: Whether to return all hidden states.
            return_dict: Whether to return a dictionary or tensor as output.

        Returns:
            If return_dict is True, returns a dictionary with model outputs.
            Otherwise, returns the output tensor.
        """
        if isinstance(inputs, (list, tuple)):
            x, encoder_feature, decoder_feature = inputs
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        elif isinstance(inputs, dict):
            x = inputs["x"]
            encoder_feature = inputs["encoder_feature"]
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        else:
            encoder_feature = x = inputs

        batch_size, _, n_feature = tf.shape(encoder_feature)
        # # de-comp
        # seasonal_init, trend_init = self.series_decomp(encoder_feature)
        # # decoder input
        # seasonal_init = tf.concat(
        #     [seasonal_init, tf.zeros([batch_size, self.predict_sequence_length, n_feature])], axis=1
        # )
        # trend_init = tf.concat(
        #     [
        #         trend_init,
        #         tf.repeat(
        #             tf.reduce_mean(encoder_feature, axis=1)[:, tf.newaxis, :],
        #             repeats=self.predict_sequence_length,
        #             axis=1,
        #         ),
        #     ],
        #     axis=1,
        # )

        # encoder
        for layer in self.encoder:
            x = layer(x)

        encoder_output = x[:, -1]
        encoder_output = self.drop1(encoder_output)
        encoder_output = self.dense1(encoder_output)
        # encoder_output = self.bn2(encoder_output)
        encoder_output = self.drop2(encoder_output)
        encoder_output = self.dense2(encoder_output)
        encoder_output = self.drop2(encoder_output)

        outputs = self.project1(encoder_output)
        outputs = tf.expand_dims(outputs, -1)
        return outputs

        # encoder_output = x
        # trend_part= trend_init[..., 0:1]

        # for layer in self.decoder:
        #     seasonal_part, trend_part = layer(seasonal_init, encoder_output, trend_part)

        # trend_part = trend_part[:, -self.predict_sequence_length:, :]
        # seasonal_part = seasonal_part[:, -self.predict_sequence_length:, :]
        # seasonal_part = self.project(seasonal_part)
        # return trend_part + seasonal_part


class EncoderLayer(tf.keras.layers.Layer):
    """Encoder Layer for Autoformer architecture.

    This layer implements the encoder block of the Autoformer architecture,
    which combines auto-correlation attention with series decomposition.

    Attributes:
        kernel_size: Size of the kernel for series decomposition.
        d_model: The dimension of the model/embedding.
        num_attention_heads: Number of attention heads.
        dropout_rate: Dropout rate for regularization.
    """

    def __init__(self, kernel_size: int, d_model: int, num_attention_heads: int, dropout_rate: float = 0.1) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.d_model = d_model
        self.num_attention_heads = num_attention_heads
        self.dropout_rate = dropout_rate

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        self.series_decomp1 = SeriesDecomp(self.kernel_size)
        self.series_decomp2 = SeriesDecomp(self.kernel_size)
        self.autocorrelation = AutoCorrelation(self.d_model, self.num_attention_heads)
        self.drop = Dropout(self.dropout_rate)
        self.dense = Dense(input_shape[-1])

    def call(self, x: tf.Tensor) -> tf.Tensor:
        """Process input through the encoder layer.

        Args:
            x: Input tensor of shape [batch_size, time_steps, features]
            training: Whether the layer is being called during training

        Returns:
            Processed tensor after applying encoder operations
        """
        x, _ = self.series_decomp1(self.autocorrelation(x, x, x) + x)
        x, _ = self.series_decomp2(self.drop(self.dense(x)) + x)
        return x


class DecoderLayer(tf.keras.layers.Layer):
    """Decoder Layer for Autoformer architecture.

    This layer implements the decoder block of the Autoformer architecture,
    which combines auto-correlation attention with series decomposition.

    Attributes:
        kernel_size: Size of the kernel for series decomposition.
        d_model: The dimension of the model/embedding.
        num_attention_heads: Number of attention heads.
        drop_rate: Dropout rate for regularization.
    """

    def __init__(self, kernel_size, d_model, num_attention_heads, drop_rate=0.1) -> None:
        super().__init__()
        self.kernel_size = kernel_size
        self.d_model = d_model
        self.num_attention_heads = num_attention_heads
        self.drop_rate = drop_rate

    def build(self, input_shape: Tuple[Optional[int], ...]) -> None:
        """Build the layer with input shape information."""
        self.series_decomp1 = SeriesDecomp(self.kernel_size, name="series_decomp1")
        self.series_decomp2 = SeriesDecomp(self.kernel_size, name="series_decomp2")
        self.series_decomp3 = SeriesDecomp(self.kernel_size, name="series_decomp3")
        self.autocorrelation1 = AutoCorrelation(self.d_model, self.num_attention_heads)
        self.autocorrelation2 = AutoCorrelation(self.d_model, self.num_attention_heads)
        self.conv1 = Conv1D(self.d_model, kernel_size=3, strides=1, padding="same")
        self.project = Conv1D(1, kernel_size=3, strides=1, padding="same")
        self.drop = Dropout(self.drop_rate)
        self.dense1 = Dense(input_shape[-1])
        self.conv2 = Conv1D(input_shape[-1], kernel_size=3, strides=1, padding="same")
        self.activation = ReLU()

    def call(self, x, cross, init_trend):
        """Process inputs through the decoder layer.

        Args:
            x: Input tensor of shape [batch_size, time_steps, features]
            cross: Cross attention tensor of shape [batch_size, time_steps, features]
            init_trend: Initial trend component of shape [batch_size, time_steps, 1]
            training: Whether the layer is being called during training

        Returns:
            Tuple of (seasonal component, updated trend component)
        """
        x, trend1 = self.series_decomp1(self.drop(self.autocorrelation1(x, x, x)) + x)
        x, trend2 = self.series_decomp2(self.drop(self.autocorrelation2(x, cross, cross)) + x)
        x = self.conv2(self.drop(self.activation(self.conv1(x))))
        x, trend3 = self.series_decomp3(self.drop(self.dense1(x)) + x)

        trend = trend1 + trend2 + trend3
        trend = self.drop(self.project(trend))
        final_trend = init_trend + trend
        return x, final_trend



"""
`RWKV: Reinventing RNNs for the Transformer Era
<https://arxiv.org/abs/2305.13048>`_
"""

from typing import Dict, Optional

import tensorflow as tf
from tensorflow.keras.layers import GRU, Dense

from tfts.layers.rwkv_layer import ChannelMixing, TimeMixing

from .base import BaseConfig, BaseModel


class RWKVConfig(BaseConfig):
    model_type: str = "rwkv"

    def __init__(
        self,
        num_layers: int = 25,
        hidden_size: int = 64,
        vocab_size: int = 50304,
        dense_hidden_size: int = 32,
        dropout: float = 0.0,
    ) -> None:
        """
        Initializes the configuration for the RWKV model with the specified parameters.

        Args:
            num_layers: The number of stacked RWKV layers.
            hidden_size: Size of each attention head.
            dense_hidden_size: The size of the dense hidden layer following the RWKV.
        """
        super().__init__()

        self.vocab_size: int = vocab_size
        self.num_layers: int = num_layers
        self.hidden_size: int = hidden_size
        self.dense_hidden_size: int = dense_hidden_size
        self.dropout: float = dropout


class RWKV(BaseModel):
    """TensorFlow RWKV model"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[RWKVConfig] = None):
        super().__init__()
        self.config = config or RWKVConfig()
        self.predict_sequence_length = predict_sequence_length

        self.emb = tf.keras.layers.Embedding(self.config.vocab_size, self.config.hidden_size)
        self.ln0 = tf.keras.layers.LayerNormalization()

        self.blocks = [RWKVBlock(self.config) for _ in range(self.config.num_layers)]

        self.ln_out = tf.keras.layers.LayerNormalization()
        self.head = tf.keras.layers.Dense(self.config.vocab_size, use_bias=False)

    def init_state(self, batch_size=1):
        states = []
        for _ in range(self.config.num_layers):
            # States for attention
            att_states = [
                tf.zeros((batch_size, self.config.hidden_size)),  # aa
                tf.zeros((batch_size, self.config.hidden_size)),  # bb
                tf.zeros((batch_size, self.config.hidden_size)) - 1e30,  # pp
            ]
            # State for FFN
            ffn_state = tf.zeros((batch_size, self.config.hidden_size))
            states.append((att_states, ffn_state))
        return states

    def __call__(
        self,
        x,
        states=None,
        teacher=None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """RWKV model call"""

        if states is None:
            states = self.init_state()

        x = self.emb(x)
        x = self.ln0(x)

        new_states = []
        for i, block in enumerate(self.blocks):
            x, block_states = block(x, states[i])
            new_states.append(block_states)

        x = self.ln_out(x)
        x = self.head(x)

        return x, new_states

    def _prepare_inputs(self, inputs):
        """Prepare the inputs for the encoder."""

        if isinstance(inputs, (list, tuple)):
            x, encoder_feature, _ = inputs
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        elif isinstance(inputs, dict):
            x = inputs["x"]
            encoder_feature = inputs["encoder_feature"]
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        else:
            x = inputs
            encoder_feature = x
        return x, encoder_feature


class RWKVBlock(tf.keras.layers.Layer):
    """TensorFlow RWKV block"""

    def __init__(self, config, **kwargs):
        super().__init__(**kwargs)
        self.ln1 = tf.keras.layers.LayerNormalization()
        self.attention = TimeMixing(config)
        self.ln2 = tf.keras.layers.LayerNormalization()
        self.feed_forward = ChannelMixing(config)

    def call(self, x, states):
        """block

        Parameters
        ----------
        x : tf.Tensor
            The input tensor of shape (batch_size, seq_length, embed_dim).
        """
        att_states, ffn_state = states

        # Attention
        h, new_att_states = self.attention(self.ln1(x), att_states)
        x = x + h

        # Feed-forward
        h, new_ffn_state = self.feed_forward(self.ln2(x), ffn_state)
        x = x + h

        return x, (new_att_states, new_ffn_state)



"""
`A decoder-only foundation model for time-series forecasting
<https://arxiv.org/abs/2310.10688>`_
"""

import logging

import tensorflow as tf

from .base import BaseConfig, BaseModel

logger = logging.getLogger(__name__)


class TimesFmConfig(BaseConfig):
    model_type: str = "timesfm"

    def __init__(self):
        super().__init__()



"""Layer for :py:class:`~tfts.models.nbeats`"""

from typing import Dict, Optional, Tuple

import numpy as np
import tensorflow as tf
from tensorflow.keras.layers import Activation, Dense


class GenericBlock(tf.keras.layers.Layer):
    """Generic block"""

    def __init__(
        self, train_sequence_length: int, predict_sequence_length: int, hidden_size: int, n_block_layers: int = 4
    ):
        super(GenericBlock, self).__init__()
        self.train_sequence_length = train_sequence_length
        self.predict_sequence_length = predict_sequence_length
        self.hidden_size = hidden_size
        self.n_block_layers = n_block_layers

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.layers = [Dense(self.hidden_size, activation="relu") for _ in range(self.n_block_layers)]
        self.theta = Dense(self.train_sequence_length + self.predict_sequence_length, use_bias=False, activation=None)
        super(GenericBlock, self).build(input_shape)

    def call(self, inputs):
        """Compute the output of the Generic Block.

        Parameters
        ----------
        inputs : tf.Tensor
            A tensor of shape (batch_size, train_sequence_length, input_size)

        Returns
        -------
        Tuple[tf.Tensor, tf.Tensor]
            A tuple of two tensors, the first of shape (batch_size, train_sequence_length, output_size)
            and the second of shape (batch_size, predict_sequence_length, output_size)
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)
        x = self.theta(x)
        return x[:, : self.train_sequence_length], x[:, -self.predict_sequence_length :]


class TrendBlock(tf.keras.layers.Layer):
    """Trend block"""

    def __init__(
        self,
        train_sequence_length: int,
        predict_sequence_length: int,
        hidden_size: int,
        n_block_layers: int = 4,
        polynomial_term=2,
    ):
        super().__init__()
        self.train_sequence_length = train_sequence_length
        self.predict_sequence_length = predict_sequence_length
        self.hidden_size = hidden_size
        self.n_bloack_layers = n_block_layers
        self.polynomial_size = polynomial_term + 1
        self.forecast_time = tf.concat(
            [
                tf.math.pow(tf.range(predict_sequence_length, dtype=tf.float32) / predict_sequence_length, i)[None, :]
                for i in range(self.polynomial_size)
            ],
            axis=0,
        )
        self.backcast_time = tf.concat(
            [
                tf.math.pow(tf.range(train_sequence_length, dtype=tf.float32) / train_sequence_length, i)[None, :]
                for i in range(self.polynomial_size)
            ],
            axis=0,
        )

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.layers = [Dense(self.hidden_size, activation="relu") for _ in range(self.n_bloack_layers)]
        self.theta = Dense(2 * self.polynomial_size, use_bias=False, activation=None)
        super().build(input_shape)

    def call(self, inputs):
        """_summary_

        Parameters
        ----------
        inputs : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)
        x = self.theta(x)
        backcast = tf.einsum("bp,pt->bt", x[:, self.polynomial_size :], self.backcast_time)
        forecast = tf.einsum("bp,pt->bt", x[:, : self.polynomial_size], self.forecast_time)
        return backcast, forecast


class SeasonalityBlock(tf.keras.layers.Layer):
    """Seasonality block"""

    def __init__(self, train_sequence_length, predict_sequence_length, hidden_size, n_block_layers=4, num_harmonics=1):
        super().__init__()
        self.train_sequence_length = train_sequence_length
        self.predict_sequence_length = predict_sequence_length
        self.hidden_size = hidden_size
        self.n_block_layers = n_block_layers
        self.num_harmonics = num_harmonics
        self.theta_size = 4 * int(np.ceil(num_harmonics / 2 * predict_sequence_length) - (num_harmonics - 1))

        self.frequency = tf.concat(
            [
                tf.zeros(1, dtype=tf.float32),
                tf.range(num_harmonics, num_harmonics / 2 * predict_sequence_length, dtype=tf.float32),
            ],
            axis=0,
        )
        self.backcast_grid = (
            -2
            * np.pi
            * (tf.range(self.train_sequence_length, dtype=tf.float32)[:, None] / train_sequence_length)
            * self.frequency
        )
        self.forecast_grid = (
            2
            * np.pi
            * (tf.range(predict_sequence_length, dtype=tf.float32)[:, None] / predict_sequence_length)
            * self.frequency
        )
        self.backcast_cos_template = tf.transpose(tf.cos(self.backcast_grid))
        self.backcast_sin_template = tf.transpose(tf.sin(self.backcast_grid))
        self.forecast_cos_template = tf.transpose(tf.cos(self.forecast_grid))
        self.forecast_sin_template = tf.transpose(tf.sin(self.forecast_grid))

    def build(self, input_shape: Tuple[Optional[int], ...]):
        self.layers = [Dense(self.hidden_size, activation="relu") for _ in range(self.n_block_layers)]
        self.theta = Dense(self.theta_size, use_bias=False, activation=None)

    def call(self, inputs):
        """_summary_

        Parameters
        ----------
        inputs : _type_
            _description_

        Returns
        -------
        _type_
            _description_
        """
        x = inputs
        for layer in self.layers:
            x = layer(x)
        x = self.theta(x)

        config_per_harmonic = self.theta_size // 4

        backcast_harmonics_cos = tf.einsum(
            "bp,pt->bt", inputs[:, 2 * config_per_harmonic : 3 * config_per_harmonic], self.backcast_cos_template
        )
        backcast_harmonics_sin = tf.einsum("bp,pt->bt", x[:, 3 * config_per_harmonic :], self.backcast_sin_template)
        backcast = backcast_harmonics_sin + backcast_harmonics_cos

        forecast_harmonics_cos = tf.einsum("bp,pt->bt", x[:, :config_per_harmonic], self.forecast_cos_template)
        forecast_harmonics_sin = tf.einsum(
            "bp,pt->bt", x[:, config_per_harmonic : 2 * config_per_harmonic], self.forecast_sin_template
        )
        forecast = forecast_harmonics_sin + forecast_harmonics_cos

        return backcast, forecast



"""tfts Generator"""


class GenerationMixin:
    """
    A class containing all functions for auto-regressive generation, to be used as a mixin.
    """

    def generate(self, inputs, generation_config):
        return



"""
`Language Models are Few-Shot Learners
<https://arxiv.org/abs/2005.14165>`_
"""

from typing import Dict, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Dense, Reshape

from tfts.layers.embed_layer import DataEmbedding, TokenEmbedding
from tfts.models.transformer import Encoder

from .base import BaseConfig, BaseModel


class GPTConfig(BaseConfig):

    model_type: str = "gpt"

    def __init__(
        self,
        hidden_size: int = 64,
        num_layers: int = 2,
        num_attention_heads: int = 4,
        ffn_intermediate_size: int = 256,
        hidden_act: str = "gelu",
        hidden_dropout_prob: float = 0.0,
        attention_probs_dropout_prob: float = 0.0,
        max_position_embeddings: int = 512,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        layer_norm_eps: float = 1e-12,
        pad_token_id: int = 0,
        positional_type: str = "absolute",
        use_cache: bool = True,
        dense_units: Tuple[int] = (512, 1024),
        classifier_dropout: Optional[float] = None,
        **kwargs: Dict[str, object],
    ) -> None:
        """Configuration class for GPT decoder model, inheriting from BaseConfig.

        Args:
            hidden_size: The size of the hidden layers. Default is 64.
            num_hidden_layers: The number of hidden layers in the transformer encoder. Default is 2.
            num_attention_heads: The number of attention heads in each attention layer. Default is 4.
            ffn_intermediate_size: The size of the intermediate (feed-forward) layer. Default is 256.
            hidden_act: The activation function for hidden layers. Default is "gelu".
            hidden_dropout_prob: The dropout probability for hidden layers. Default is 0.1.
            attention_probs_dropout_prob: The dropout probability for attention probabilities. Default is 0.1.
            max_position_embeddings: The maximum length of the input sequences. Default is 512.
            type_vocab_size: The vocabulary size for token types (usually 2). Default is 2.
            initializer_range: The standard deviation for weight initialization. Default is 0.02.
            layer_norm_eps: The epsilon value for layer normalization. Default is 1e-12.
            pad_token_id: The ID for the padding token. Default is 0.
            positional_type: The type of position embedding ("absolute" or "relative"). Default is "absolute".
            use_cache: Whether to use the cache during inference. Default is True.
            classifier_dropout: Dropout probability for the classifier layer. Default is None.
            **kwargs: Additional keyword arguments passed to the parent `BaseConfig` class.
        """

        super().__init__(**kwargs)

        self.hidden_size: int = hidden_size
        self.num_layers: int = num_layers
        self.num_attention_heads: int = num_attention_heads
        self.ffn_intermediate_size: int = ffn_intermediate_size
        self.hidden_act: str = hidden_act
        self.hidden_dropout_prob: float = hidden_dropout_prob
        self.attention_probs_dropout_prob: float = attention_probs_dropout_prob
        self.max_position_embeddings: int = max_position_embeddings
        self.type_vocab_size: int = type_vocab_size
        self.initializer_range: float = initializer_range
        self.layer_norm_eps: float = layer_norm_eps
        self.positional_type: str = positional_type
        self.use_cache: bool = use_cache
        self.dense_unites: Tuple[int] = dense_units
        self.classifier_dropout: Optional[float] = classifier_dropout
        self.pad_token_id: int = pad_token_id

    def __post_init__(self):
        """Validate configuration parameters."""
        if self.hidden_size <= 0:
            raise ValueError(f"hidden_size must be positive, got {self.hidden_size}")
        if self.num_layers <= 0:
            raise ValueError(f"num_layers must be positive, got {self.num_layers}")
        if self.num_attention_heads <= 0:
            raise ValueError(f"num_attention_heads must be positive, got {self.num_attention_heads}")
        if not 0 <= self.attention_probs_dropout_prob < 1:
            raise ValueError(f"attention_probs_dropout_prob must be in [0, 1), got {self.attention_probs_dropout_prob}")
        if not 0 <= self.hidden_dropout_prob < 1:
            raise ValueError(f"hidden_dropout_prob must be in [0, 1), got {self.hidden_dropout_prob}")
        if self.hidden_size % self.num_attention_heads != 0:
            raise ValueError(
                f"hidden_size must be divisible by attention_heads, got {self.hidden_size}/{self.num_attention_heads}"
            )


class GPT(BaseModel):
    """GPT decoder model for time series"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[GPTConfig] = None) -> None:
        super(GPT, self).__init__()
        self.config = config or GPTConfig()
        self.predict_sequence_length = predict_sequence_length

        self.encoder_embedding = TokenEmbedding(self.config.hidden_size)
        self.encoder = Encoder(
            num_hidden_layers=self.config.num_layers,
            hidden_size=self.config.hidden_size,
            num_attention_heads=self.config.num_attention_heads,
            attention_probs_dropout_prob=self.config.attention_probs_dropout_prob,
            ffn_intermediate_size=self.config.ffn_intermediate_size,
            hidden_dropout_prob=self.config.hidden_dropout_prob,
        )

        self.dense_layers = []
        for unit in self.config.dense_unites:
            self.dense_layers.append(Dense(unit, activation="relu"))

        self.projection = Dense(predict_sequence_length, activation=None)

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> tf.Tensor:
        """GPT model forward pass.

        Args:
            inputs: Input time-series data, can be:
                - A single tensor of shape [batch_size, seq_len, feature_dim]
                - A tuple/list of (x, encoder_feature, decoder_feature)
                - A dictionary with keys 'x' and 'encoder_feature'
            teacher: Optional teacher forcing tensor for autoregression.
            training: Whether the model is in training mode.
            mask: Optional attention mask.
            output_hidden_states: Whether to return hidden states.
            return_dict: Whether to return outputs as a dictionary.

        Returns:
            If return_dict is False and output_hidden_states is False:
                Forecasted values tensor of shape [batch_size, predict_sequence_length, input_dim]
            If output_hidden_states is True:
                Hidden states from the encoder.
            If return_dict is True:
                Dictionary containing model outputs.
        """
        if isinstance(inputs, (list, tuple)):
            x, encoder_feature, decoder_feature = inputs
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        elif isinstance(inputs, dict):
            x = inputs["x"]
            encoder_feature = inputs["encoder_feature"]
            encoder_feature = tf.concat([x, encoder_feature], axis=-1)
        else:
            encoder_feature = x = inputs

        encoder_feature = self.encoder_embedding(encoder_feature)

        memory = self.encoder(encoder_feature, mask=None)

        if output_hidden_states:
            # (batch_size, train_sequence_length, hidden_size)
            return memory

        encoder_output = memory[:, -1]

        for layer in self.dense_layers:
            encoder_output = layer(encoder_output)
        outputs = self.projection(encoder_output)
        outputs = Reshape((outputs.shape[1], 1))(outputs)
        return outputs



"""
`WaveNet: A Generative Model for Raw Audio
<https://arxiv.org/abs/1609.03499>`_
"""

from typing import List, Optional, Tuple

import tensorflow as tf
from tensorflow.keras.layers import Concatenate, Conv1D, Dense, Dropout, Lambda, ReLU, Reshape

from tfts.layers.cnn_layer import ConvTemp
from tfts.layers.dense_layer import DenseTemp

from .base import BaseConfig, BaseModel


class TCNConfig(BaseConfig):
    model_type: str = "tcn"

    def __init__(
        self,
        dilation_rates: List[int] = [2**i for i in range(4)],
        kernel_sizes: List[int] = [2 for _ in range(4)],
        filters: int = 128,
        dense_hidden_size: int = 64,
    ):
        """
        Initializes the configuration for the Temporal Convolutional Network (TCN) model with the specified parameters.

        Args:
            dilation_rates: List of dilation rates for each layer.
            kernel_sizes: List of kernel sizes for each convolutional layer.
            filters: The number of filters (channels) in each convolutional layer.
            dense_hidden_size: The size of the dense hidden layer.
        """
        super().__init__()
        self.dilation_rates: List[int] = dilation_rates
        self.kernel_sizes: List[int] = kernel_sizes
        self.filters: int = filters
        self.dense_hidden_size: int = dense_hidden_size


class TCN(BaseModel):
    """Temporal convolutional network"""

    def __init__(self, predict_sequence_length: int = 1, config: Optional[TCNConfig] = None) -> None:
        super(TCN, self).__init__()
        self.config = config or TCNConfig()
        self.predict_sequence_length = predict_sequence_length
        self.encoder = Encoder(
            kernel_sizes=self.config.kernel_sizes,
            dilation_rates=self.config.dilation_rates,
            filters=self.config.filters,
            dense_hidden_size=self.config.dense_hidden_size,
        )

        self.project1 = Dense(predict_sequence_length, activation=None)

        self.drop1 = Dropout(0.25)
        self.dense1 = Dense(512, activation="relu")

        self.drop2 = Dropout(0.25)
        self.dense2 = Dense(1024, activation="relu")

    def __call__(
        self,
        inputs: tf.Tensor,
        teacher: Optional[tf.Tensor] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ):
        """TCN call

        Parameters
        ----------
        inputs : tf.Tensor
            3D input tensor
        teacher : tf.Tensor, optional
            _description_, by default None

        Returns
        -------
        tf.Tensor
            3D output tensor
        """
        x, encoder_feature, _ = self._prepare_3d_inputs(inputs)

        # encoder_features = self.pool(encoder_features)  # batch * n_train_days * n_feature
        encoder_outputs, encoder_state = self.encoder(encoder_feature)
        # outputs = self.dense1(encoder_state)  # batch * predict_sequence_length
        # outputs = self.dense2(encoder_outputs)[:, -self.predict_sequence_length]

        if output_hidden_states:
            return encoder_outputs

        memory = encoder_state[:, -1]
        encoder_output = self.drop1(memory)
        encoder_output = self.dense1(encoder_output)
        encoder_output = self.drop2(encoder_output)
        encoder_output = self.dense2(encoder_output)
        encoder_output = self.drop2(encoder_output)

        outputs = self.project1(encoder_output)
        outputs = Reshape((outputs.shape[1], 1))(outputs)

        # outputs = tf.tile(outputs, (1, self.predict_sequence_length, 1))   # stupid
        # outputs = self.dense3(encoder_outputs)
        return outputs


class Encoder(object):
    def __init__(self, kernel_sizes, dilation_rates, filters, dense_hidden_size):
        self.filters = filters
        self.conv_times = []
        for i, (kernel_size, dilation) in enumerate(zip(kernel_sizes, dilation_rates)):
            self.conv_times.append(
                ConvTemp(filters=2 * filters, kernel_size=kernel_size, causal=True, dilation_rate=dilation)
            )
        self.dense_time1 = DenseTemp(hidden_size=filters, activation="tanh", name="encoder_dense_time1")
        self.dense_time2 = DenseTemp(hidden_size=filters + filters, name="encoder_dense_time2")
        self.dense_time3 = DenseTemp(hidden_size=dense_hidden_size, activation="relu", name="encoder_dense_time3")
        self.dense_time4 = DenseTemp(hidden_size=1, name="encoder_dense_time_4")

    def __call__(self, x: tf.Tensor):
        # => batch_size * time_sequence_length * filters
        inputs = self.dense_time1(inputs=x)

        skip_outputs = []
        conv_inputs = [inputs]
        for conv_time in self.conv_times:
            dilated_conv = conv_time(inputs)

            split_layer = Lambda(lambda x: tf.split(x, 2, axis=2))
            conv_filter, conv_gate = split_layer(dilated_conv)
            # dilated_conv = tf.nn.tanh(conv_filter) * tf.nn.sigmoid(conv_gate)
            dilated_conv = Lambda(lambda x: tf.nn.tanh(x[0]) * tf.nn.sigmoid(x[1]))([conv_filter, conv_gate])
            outputs = self.dense_time2(inputs=dilated_conv)
            # skips, residuals = tf.split(outputs, [self.filters, self.filters], axis=2)
            split_layer2 = Lambda(lambda x: tf.split(x, [self.filters, self.filters], axis=2))
            skips, residuals = split_layer2(outputs)
            inputs += residuals
            conv_inputs.append(inputs)  # batch_size * time_sequence_length * filters
            skip_outputs.append(skips)

        # skip_outputs = tf.nn.relu(tf.concat(skip_outputs, axis=2))
        concat_layer = Concatenate(axis=2)
        concatenated = concat_layer(skip_outputs)
        relu_layer = ReLU()
        skip_outputs = relu_layer(concatenated)
        h = self.dense_time3(skip_outputs)
        # y_hat = self.dense_time4(h)
        return conv_inputs[:-1], h



"""Layer for :py:class:`~tfts.models.transformer`"""

import tensorflow as tf
from tensorflow.keras import activations, constraints, initializers, regularizers


class CausalMask:
    """Casual Mask is used for transformer decoder, used in first self-attention for decoder feature"""

    def __init__(self, B, L):
        mask_shape = [B, L, L]  # for multi-heads split [B, 1, L, L]
        mask_a = tf.linalg.band_part(tf.ones(mask_shape), 0, -1)  # Upper triangular matrix of 0s and 1s
        mask_b = tf.linalg.band_part(tf.ones(mask_shape), 0, 0)  # Diagonal matrix of 0s and 1s
        mask = tf.cast(mask_a - mask_b, dtype=tf.float32)

        self._mask = mask
        tf.stop_gradient(self._mask)

    @property
    def mask(self):
        return self._mask


class ProbMask:
    """ProbMask for informer"""

    def __init__(self, B, H, L, index, scores):
        # B: batch_size, H: num_attention_heads, L: seq_length
        mask = tf.ones([L, scores.shape[-1]], tf.float32)

        mask = 1 - tf.linalg.band_part(mask, -1, 0)
        mask_expanded = tf.broadcast_to(mask, [B, H, L, scores.shape[-1]])
        # mask specific q based on reduced Q
        mask_Q = tf.gather_nd(mask_expanded, index)
        self._mask = tf.cast(tf.reshape(mask_Q, tf.shape(scores)), tf.bool)

    @property
    def mask(self):
        return self._mask
